import sys
import logging
import warnings
import re
import time
import uuid
from weakref import ref

from weakref import WeakSet

from .. import _p4p
from .._p4p import (Server as _Server,
                    StaticProvider as _StaticProvider,
                    DynamicProvider as _DynamicProvider,
                    ServerOperation,
                    )

if sys.version_info >= (3, 0):
    unicode = str

_log = logging.getLogger(__name__)

__all__ = (
    'Server',
        'installProvider',
        'removeProvider',
        'StaticProvider',
        'DynamicProvider',
        'ServerOperation',
)

def installProvider(name, provider):
    _p4p._providers[name] = ref(provider)
def removeProvider(name):
    _p4p._providers.pop(name, None)
def clearProviders():
    _p4p._providers.clear()

class Server(object):

    """Server(conf=None, useenv=True, providers=[""])

    :param providers: A list of provider names or instances.  See below.
    :param dict conf: Configuration keys for the server.  Uses same names as environment variables (aka. EPICS_PVAS_*)
    :param bool useenv: Whether to use process environment in addition to provided config.
    :param bool isolate: If True, override conf= and useenv= to select a configuration suitable for isolated testing.
                         eg. listening only on localhost with a randomly chosen port number.  Use `conf()` to determine
                         which port is being used.

    Run a PVAccess server serving Channels from the listed providers.
    The server is running after construction, until stop(). ::

        S = Server(providers=["example"])
        # do something else
        S.stop()

    As a convenience, a Server may be used as a context manager to automatically `stop()`. ::

        with Server(providers=["example"]) as S:
            # do something else

    When configuring a Server, conf keys provided to the constructor have the same name as the environment variables.
    If both are given, then the provided conf dict is used.

    Call Server.conf() to see a list of valid server (EPICS_PVAS_*) key names and the actual values.

    The providers list may contain: name strings (cf. installProvider()),
    `StaticProvider` or `DynamicProvider` instances, or a dict "{'pv:name':`SharedPV`}" to implicitly creat a `StaticProvider`.
    Each entry may also be a tuple "(provider, order)" where "provider" is any of the allowed types,
    and "order" is an integer used to resolve ambiguity if more than one provider may claim a PV name.
    (lower numbers are queried first, the default order is 0)
    """

    def __init__(self, providers, isolate=False, **kws):
        self.__keep_alive = [] # ick...

        if isinstance(providers, (bytes, unicode)):
            providers = providers.split() # split on space
            warnings.warn("Server providers list should be a list", DeprecationWarning)

        Ps = []
        for provider in providers:
            if isinstance(provider, tuple):
                provider, order = provider
            elif hasattr(provider, 'order'):
                order = provider.order
            else:
                order = 0

            if isinstance(provider, (bytes, unicode)):
                if not re.match(r'^[^ \t\n\r]+$', provider):
                    raise ValueError("Invalid provider name: '%s'"%provider)
                Ps.append((provider, order))

            elif isinstance(provider, (_StaticProvider, _DynamicProvider, _p4p.Source)):
                Ps.append((provider, order))

            elif hasattr(provider, 'items'):
                P = StaticProvider()
                for name, pv in provider.items():
                    P.add(name, pv)
                Ps.append((P, order))
                # Normally user code is responsible for keeping the StaticProvider alive.
                # Not possible in this case though.
                self.__keep_alive.append(P)

            else:
                raise ValueError("providers=[] must be a list of string, SharedPV, or dict.  Not %s"%provider)

        if isolate:
            assert 'useenv' not in kws and 'conf' not in kws, kws
            kws['useenv'] = False
            kws['conf'] = {
                'EPICS_PVAS_INTF_ADDR_LIST': '127.0.0.1',
                'EPICS_PVA_ADDR_LIST': '127.0.0.1',
                'EPICS_PVA_AUTO_ADDR_LIST': '0',
                'EPICS_PVA_SERVER_PORT': '0',
                'EPICS_PVA_BROADCAST_PORT': '0',
            }


        _log.debug("Starting Server isolated=%s, %s", isolate, kws)
        self._S = _Server(providers=Ps, **kws)

        self.tostr = self._S.tostr

        self._S.start()
        try:
            if _log.isEnabledFor(logging.DEBUG):
                _log.debug("New Server: %s", self.tostr(5))
        except:
            self._S.stop()
            raise

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.stop()

    def conf(self):
        """Return a dict() with the effective configuration this server is using.

        Suitable to pass to another Server to duplicate this configuration,
        or to a client Context to allow it to connect to this server. ::

            with Server(providers=["..."], isolate=True) as S:
                with p4p.client.thread.Context('pva', conf=S.conf(), useenv=False) as C:
                    print(C.get("pv:name"))
        """
        return self._S.conf()

    def stop(self):
        """Force server to stop serving, and close connections to existing clients.
        """
        _log.debug("Stopping Server")
        self._S.stop()
        self.__keep_alive = []

    @classmethod
    def forever(klass, *args, **kws):
        """Create a server and block the calling thread until KeyboardInterrupt.
        Shorthand for: ::

            with Server(*args, **kws):
                try;
                    time.sleep(99999999)
                except KeyboardInterrupt:
                    pass
        """
        with klass(*args, **kws):
            _log.info("Running server")
            try:
                while True:
                    time.sleep(100)
            except KeyboardInterrupt:
                pass
            finally:
                _log.info("Stopping server")

class StaticProvider(_StaticProvider):

    """A channel provider which servers from a clearly defined list of names.
    This list may change at any time.

    :param str name: Provider name.  Must be unique within the local context in which it is used.
                     None, the default, will choose an appropriate value.
    """
    def __init__(self, name=None):
        if name is None:
            # Caller doesn't care.  Pick something unique w/o spaces
            name = str(uuid.uuid4())
        super(StaticProvider, self).__init__(name)

class DynamicProvider(_DynamicProvider):

    """A channel provider which does not maintain a list of provided channel names.

       The following example shows a simple case, in fact so simple that StaticProvider
       is a better fit. ::

            class DynHandler(object):
                def __init__(self):
                    self.pv = SharedPV()
                def testChannel(self, name): # return True, False, or DynamicProvider.NotYet
                    return name=="blah"
                def makeChannel(self, name, peer):
                    assert name=="blah"
                    return self.pv
            provider = DynamicProvider("arbitrary", DynHandler())
            server = Server(providers=[provider])
    """

    # Return from Handler.testChannel() to prevent caching of negative result.
    # Use when testChannel('name') might shortly return True
    NotYet = b'nocache'

    def __init__(self, name, handler):
        _DynamicProvider.__init__(self, name, self._WrapHandler(handler))

    class _WrapHandler(object):

        "Wrapper around user Handler which logs exception"

        def __init__(self, real):
            self._real = real

        def testChannel(self, name):
            try:
                return self._real.testChannel(name)
            except:
                _log.exception("Unexpected")

        def makeChannel(self, name, peer):
            try:
                return self._real.makeChannel(name, peer)
            except:
                _log.exception("Unexpected")

def _cleanup_servers():
    _log.debug("Stopping all Server instances")
    servers = list(_p4p.all_servers)
    for srv in servers:
        srv.stop()




import logging
_log = logging.getLogger(__name__)

try:
    from itertools import izip
except ImportError:
    izip = zip

from collections import OrderedDict
from ..wrapper import Type, Value
from .common import timeStamp, alarm, NTBase
from .scalar import NTScalar
from .ndarray import NTNDArray
from .enum import NTEnum

__all__ = [
    'NTScalar',
    'NTEnum',
    'NTMultiChannel',
    'NTTable',
    'NTNDArray',
    'defaultNT',
]

_default_nt = {
    "epics:nt/NTScalar:1.0": NTScalar,
    "epics:nt/NTScalarArray:1.0": NTScalar,
    "epics:nt/NTEnum:1.0": NTEnum,
    "epics:nt/NTNDArray:1.0": NTNDArray,
}

def defaultNT():
    """Returns a copy of the default NT helper mappings.

    :since: 3.1.0
    """
    return _default_nt.copy()

class UnwrapOnly(object):
    def __init__(self, unwrap):
        self.unwrap = unwrap
    def __call__(self):
        return self # we are state-less
    def wrap(self, V):
        return V

def buildNT(nt=None, unwrap=None):
    if unwrap is False or nt is False:
        return ClientUnwrapper({}) # disable use of wrappers

    if unwrap is not None:
        # legacy
        ret = {} # ignore new style
        for ID,fn in (unwrap or {}).items():
            ret[ID] = UnwrapOnly(fn)

    else:
        ret = dict(_default_nt)
        ret.update(nt or {})

    return ClientUnwrapper(ret)

class ClientUnwrapper(object):
    def __init__(self, nt=None):
        self.nt = nt
        self.id = None
        self._wrap = self._unwrap = lambda x:x
        self._assign = self._default_assign
    def wrap(self, val, **kws):
        """Pack a arbitrary python object into a Value
        """
        return self._wrap(val, **kws)
    def unwrap(self, val):
        """Unpack a Value as some other python type
        """
        if val.getID()!=self.id:
            self._update(val)
        return self._unwrap(val)

    def assign(self, V, value):
        if V.getID()!=self.id:
            self._update(V)
        self._assign(V, value)

    def _update(self, val):
        # type change
        nt = self.nt.get(val.getID())
        if nt is not None:
            nt = nt() # instancate
            self._wrap, self._unwrap = nt.wrap, nt.unwrap
            self._assign = nt.assign
            self.id = val.getID()
        else:
            # reset
            self._wrap = self._unwrap = lambda x:x
            self._assign = self._default_assign

    def _default_assign(self, V, value):
        V.value = value # assume NTScalar-like

    def __repr__(self):
        return '%s(%s)'%(self.__class__.__name__, repr(self.nt))
    __str__ = __repr__

class NTMultiChannel(NTBase):

    """Describes a structure holding the equivalent of a number of NTScalar
    """
    Value = Value

    @staticmethod
    def buildType(valtype, extra=[]):
        """Build a Type

        :param str valtype: A type code to be used with the 'value' field.  Must be an array
        :param list extra: A list of tuples describing additional non-standard fields
        :returns: A :py:class:`Type`
        """
        assert valtype[:1] == 'a', 'valtype must be an array'
        return Type(id="epics:nt/NTMultiChannel:1.0",
                    spec=[
                    ('value', valtype),
                    ('channelName', 'as'),
                    ('descriptor', 's'),
                    ('alarm', alarm),
                    ('timeStamp', timeStamp),
                    ('severity', 'ai'),
                    ('status', 'ai'),
                    ('message', 'as'),
                    ('secondsPastEpoch', 'al'),
                    ('nanoseconds', 'ai'),
                    ('userTag', 'ai'),
                    ('isConnected', 'a?'),
                    ] + extra)


class NTTable(NTBase):

    """A generic table

    >>> table = NTTable.buildType(columns=[
        ('columnA', 'ai'),
        ('columnB', 'as'),
    ])
    """
    Value = Value

    @staticmethod
    def buildType(columns=[], extra=[]):
        """Build a table

        :param list columns: List of column names and types. eg [('colA', 'd')]
        :param list extra: A list of tuples describing additional non-standard fields
        :returns: A :py:class:`Type`
        """
        return Type(id="epics:nt/NTTable:1.0",
                    spec=[
                    ('labels', 'as'),
                    ('value', ('S', None, columns)),
                    ('descriptor', 's'),
                    ('alarm', alarm),
                    ('timeStamp', timeStamp),
                    ] + extra)

    def __init__(self, columns=[], extra=[]):
        self.labels = []
        C = []
        for col, type in columns:
            if type[0] == 'a':
                raise ValueError("NTTable column types may not be array")
            C.append((col, 'a' + type))
            self.labels.append(col)
        self.type = self.buildType(C, extra=extra)

    def wrap(self, values, **kws):
        """Pack an iterable of dict into a Value

        >>> T=NTTable([('A', 'ai'), ('B', 'as')])
        >>> V = T.wrap([
            {'A':42, 'B':'one'},
            {'A':43, 'B':'two'},
        ])
        """
        if isinstance(values, Value):
            return values
        else:
            cols = dict([(L, []) for L in self.labels])
            try:
                # unzip list of dict
                for V in values:
                    for L in self.labels:
                        try:
                            cols[L].append(V[L])
                        except (IndexError, KeyError):
                            pass
                # allow omit empty columns
                for L in self.labels:
                    V = cols[L]
                    if len(V) == 0:
                        del cols[L]

                try:
                    values = self.Value(self.type, {
                        'labels': self.labels,
                        'value': cols,
                    })
                except:
                    _log.error("Failed to encode '%s' with %s", cols, self.labels)
                    raise
            except:
                _log.exception("Failed to wrap: %s", values)
                raise
        return self._annotate(values, **kws)

    @staticmethod
    def unwrap(value):
        """Iterate an NTTable

        :returns: An iterator yielding an OrderedDict for each column
        """
        ret = []

        # build lists of column names, and value
        lbl, cols = [], []
        for cname, cval in value.value.items():
            lbl.append(cname)
            cols.append(cval)

        # zip together column arrays to iterate over rows
        for rval in izip(*cols):
            # zip together column names and row values
            ret.append(OrderedDict(zip(lbl, rval)))

        return ret

class NTURI(object):

    @staticmethod
    def buildType(args):
        """Build NTURI

        :param list args: A list of tuples of query argument name and PVD type code.

        >>> I = NTURI([
            ('arg_a', 'I'),
            ('arg_two', 's'),
        ])
        """
        try:
            return Type(id="epics:nt/NTURI:1.0", spec=[
                ('scheme', 's'),
                ('authority', 's'),
                ('path', 's'),
                ('query', ('S', None, args)),
            ])
        except Exception as e:
            raise ValueError('Unable to build NTURI compatible type from %s' % args)

    def __init__(self, args):
        self._args = list(args)
        self.type = self.buildType(args)

    def wrap(self, path, args=(), kws={}, scheme='', authority=''):
        """Wrap argument values (tuple/list with optional dict) into Value

        :param str path: The PV name to which this call is made
        :param tuple args: Ordered arguments
        :param dict kws: Keyword arguments
        :rtype: Value
        """
        # build dict of argument name+value
        AV = {}
        AV.update([A for A in kws.items() if A[1] is not None])
        AV.update([(N, V) for (N, _T), V in zip(self._args, args)])

        # list of argument name+type tuples for which a value was provided
        AT = [A for A in self._args if A[0] in AV]

        T = self.buildType(AT)
        try:
            return Value(T, {
                'scheme': scheme,
                'authority': authority,
                'path': path,
                'query': AV,
            })
        except Exception as e:
            raise ValueError('Unable to initialize NTURI %s from %s using %s' % (AT, AV, self._args))



# distutils: language = c++
#cython: language_level=2

cimport cython

from libc.stdint cimport uint64_t, int64_t
from libc.string cimport strcpy
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.map cimport map as cmap
from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
from cpython cimport PyObject
from cpython.bytes cimport PyBytes_FromStringAndSize

from pvxs cimport version as _version
from pvxs cimport log
from pvxs cimport util
from pvxs cimport sharedArray
from pvxs cimport data
from pvxs cimport nt
from pvxs cimport client
from pvxs cimport server
from pvxs cimport source
from pvxs cimport sharedpv

from weakref import WeakSet

cdef extern from "<p4p.h>" namespace "p4p":
    # p4p.h redefines/overrides some definitions from Python.h (eg. PyMODINIT_FUNC)
    # it also (re)defines macros effecting numpy/arrayobject.h

    # pvxs_type.cpp
    data.TypeDef startPrototype(const string& id, const data.Value& base) except+
    void appendPrototype(data.TypeDef&, object spec) except+
    object asPySpec(const data.Value& v) except+

    # pvxs_value.cpp
    int except_map()
    object asPy(const data.Value& v, bool unpackstruct, bool unpackrecurse, object wrapper) except+
    void storePy(data.Value& v, object py) except+ except_map
    object tostr(const data.Value& v, size_t limit, bool showval) except+

    # pvxs_sharedpv.cpp
    string toString(const server.Server& serv, int detail) nogil except+
    void attachHandler(sharedpv.SharedPV& pv, object handler) except+
    void detachHandler(sharedpv.SharedPV& pv) except+
    void attachCleanup(const shared_ptr[source.ExecOp]& op, object handler) except+
    void detachCleanup(const shared_ptr[source.ExecOp]& op) except+

    # pvxs_source.cpp
    shared_ptr[server.Source] createDynamic(object) except+
    void disconnectDynamic(const shared_ptr[server.Source]& src) except+

    # pvxs_client.cpp
    void opHandler[Builder](Builder& builder, object handler)
    void opBuilder[Builder](Builder& builder, object handler)
    void opEvent(client.MonitorBuilder& builder, object handler)
    object monPop(const shared_ptr[client.Subscription]& mon) with gil

cimport numpy # must cimport after p4p.h is included

numpy.import_array()
log.logger_config_env()

if not _version.version_abi_check():
    raise RuntimeError("PVXS ABI mismatch")

############### version

def version():
    return _version.version_int()

def version_str():
    return _version.version_str().decode('utf-8')

############### log

logLevelAll = <int>log.Debug
logLevelTrace = <int>log.Debug
logLevelDebug = <int>log.Debug
logLevelInfo = <int>log.Info
logLevelWarn = <int>log.Warn
logLevelError = <int>log.Err
logLevelFatal = <int>log.Crit
logLevelOff = <int>log.Crit

def logger_level_set(basestring name, int lvl):
    log.logger_level_set(name.encode(), lvl)

############### util

def listRefs():
    ret = {}
    for P in util.instanceSnapshot():
        ret[P.first.decode('utf-8')] = P.second
    return ret

def _forceLazy():
    pass

############### data

# py object to hold ownership
cdef class SharedArray:
    cdef sharedArray.shared_array[const void] arr

cdef public:
    cdef pvxs_wrap_array(const sharedArray.shared_array[const void]& arr):
        cdef SharedArray wrap = SharedArray.__new__(SharedArray)
        wrap.arr = arr
        return wrap

cdef lookupMember(data.Value* dest, const data.Value& top, key, int err):
        cdef string ckey
        cdef data.Value mem

        if key is None:
            dest[0] = top
            return
        elif isinstance(key, unicode):
            ckey = key.encode()
        else:
            ckey = key

        try:
            dest[0] = top.lookup(ckey)
        except KeyError as e:
            if err==0:
                return
            elif err==1:
                raise
            else:
                raise AttributeError(key)

cdef class _Value:
    cdef data.Value val

    def __init__(self, type=None, value=None, clone=None):
        cdef _Type base
        cdef _Value val

        if self.val.valid():
            pass

        elif type is not None:
            base = type
            self.val = base.proto.cloneEmpty()
            if value is not None:
                storePy(self.val, value)

        elif clone is not None:
            val = clone
            self.val = val.val.clone()

        else:
            raise ValueError("type= or clone= required")

    def get(self, key, default=None):
        """get(key : str, default=None) -> Value | Any
        dict-like access to sub-field

        :param str key: Sub-field name
        :param default: returned if sub-field doesn't exist
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 0)
        if not mem.valid():
            return default

        return asPy(mem, False, False, None)

    def items(self, key=None):
        """items(key : str = None) -> Iterable[Value | Any]

        :param str key: Sub-field name
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 1)

        return asPy(mem, True, False, None)

    def __setitem__(self, key, value):
        """__setitem__(key : str, value)

        :param str key: Sub-field name
        :param value: value to assign
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 1)

        storePy(mem, value)

    def __getitem__(self, key):
        """items(key : str) -> Value | Any

        :param str key: Sub-field name
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 1)

        return asPy(mem, False, False, None)

    def __getattr__(self, key):
        """__getattr__(key : str) -> Value | Any

        :param str key: Sub-field name
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 2)

        return asPy(mem, False, False, None)

    def __setattr__(self, key, value):
        """__setattr__(key : str, value)

        :param str key: Sub-field name
        :param value: value to assign
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, key, 2)

        storePy(mem, value)

    def has(self, basestring name):
        """has(name : str) -> bool
        Test for sub-field existance

        :param str name: Sub-field name
        """
        cdef string cname = name.encode()
        return self.val[cname].valid()

    def tolist(self, name=None):
        """tolist(name=None) -> List[Tuple[str, Value]]
        Return this Value (or the named sub-field) translated into a list of tuples
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, name, 2)

        return asPy(mem, True, True, None)

    def todict(self, name=None, wrapper=None):
        """todict(name=None, wrapper=None) -> Mapping[str, Value]

        Return this Value (or the named sub-field) translated into a dict

        :param str name: Sub-field name, or None
        :param callable wrapper: Passed an iterable of name,value tuples.  By default ``dict``  eg. could be OrderedDict
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, name, 2)

        return asPy(mem, True, True, wrapper or dict)

    def getID(self):
        """getID() -> str
        Return Type id= string
        """
        return self.val.id().decode() or 'structure'

    def __iter__(self):
        ret = []
        for child in self.val.ichildren():
            ret.append(self.val.nameOf(child))
        return iter(ret)

    def type(self, fld=None):
        """type(fld : str = None) -> Type
        Return the Type of this Value, or the named sub-field.

        :param str fld: Sub-field name, or None
        """
        cdef data.Value mem
        cdef _Type type
        lookupMember(&mem, self.val, fld, 1)

        type = Type.__new__(Type)
        type.proto = mem.cloneEmpty()
        return type

    def select(self, basestring name, basestring selector=None):
        """select(name : str, selector : str)
        Explicitly select Union member

        :param str name: Sub-field name
        """
        cdef string cname = name
        cdef data.Value u

        if self.val.type().code!=data.Union:
            raise KeyError(name)

        # selection is a side-effect of union lookup
        if selector is not None:
            u = self.val[cname]
        else:
            u._from(NULL);

    def changed(self, field=None):
        cdef data.Value mem
        lookupMember(&mem, self.val, field, 1)

        return mem.isMarked(True, True)

    def changedSet(self, expand=False, parents=False):
        cdef bool exp = expand
        cdef bool par = parents
        cdef data.Value parent

        ret = set()

        for mem in self.val.imarked():
            ret.add(self.val.nameOf(mem).decode())

            if exp and mem.type().code==data.Struct:
                for smem in mem.iall():
                    ret.add(self.val.nameOf(smem).decode())

            if par:
                parent = mem[b"<"]
                while parent.valid() and not self.val.equalInst(parent):
                    ret.add(self.val.nameOf(parent).decode())
                    parent = parent[b"<"]

        return ret

    def mark(self, field=None, val=True):
        """mark(field=None, val=True)
        Mark (or unmark) the this field, or the named sub-field.

        :param str field: Sub-field name
        :param bool val: To mark, or unmark
        """
        cdef data.Value mem
        lookupMember(&mem, self.val, field, 1)

        mem.mark(val)

    def unmark(self):
        """Unmark Value and all sub-fields.
        """
        self.val.unmark(True, True)

    def tostr(self, int limit=0):
        """tostr(limit : int = 0) -> str

        Return a string representation, optionally truncated to a length limit

        :param int limit: If greater than zero, formatting is terminated at ``limit`` charactors.
        """
        return tostr(self.val, limit, True)
        

cdef public api:
    cdef bool pvxs_isValue(object v):
        return isinstance(v, _Value)

    cdef data.Value pvxs_extract(object v):
        if isinstance(v, _Value):
            return (<_Value>v).val
        elif isinstance(v, _Type):
            return (<_Type?>v).proto
        else:
            return data.Value()

    cdef object pvxs_pack(const data.Value& v):
        cdef _Value raw = Value.__new__(Value)
        raw.val = v
        return raw

cdef class _Type:
    cdef data.Value proto

    def __init__(self, spec, basestring id=None, base=None):
        cdef data.TypeDef tdef
        cdef data.Value cbase
        cdef string cid

        if id is not None and base is not None:
            raise ValueError("only pass one of id= or base=")
        if id is not None:
            cid = id.encode()

        if base is not None:
            cbase = (<_Type?>base).proto

        tdef = startPrototype(cid, cbase)
        appendPrototype(tdef, spec)
        self.proto = tdef.create()

    def getID(self):
        """getId() -> str
        Return Type id= string
        """
        return self.proto.id().decode() or 'structure'

    def keys(self):
        """keys() -> Iterable[str]
        Return child field names
        """
        ret = []
        for child in self.proto.ichildren():
            ret.append(self.proto.nameOf(child).decode())
        return ret

    def aspy(self, basestring name=None):
        """aspy(str=None) -> list
        Return a Type specification list equivalent to the one passed to the constructor.
        """
        cdef string cname
        if name is None:
            return asPySpec(self.proto)
        else:
            cname = name.encode()
            return asPySpec(self.proto.lookup(cname))

    def has(self, basestring name):
        """has(str) -> bool
        Does this Type include the named member field?
        """
        cdef string cname = name.encode()
        return self.proto[cname].valid()

    def __getitem__(self, name):
        cdef string cname = name.encode()
        cdef data.Value fld = self.proto.lookup(cname)
        cdef _Type ret
        if not fld.valid():
            raise KeyError(name)
        elif fld.type().code==data.Struct:
            ret = Type.__new__(Type)
            ret.proto = fld
            return ret
        else:
            return asPySpec(fld)

    def __len__(self):
        return self.proto.nmembers()

    def tostr(self, int limit=0):
        """tostr(limit : int = 0) -> str

        Return a string representation, optionally truncated to a length limit

        :param int limit: If greater than zero, formatting is terminated at ``limit`` charactors.
        """
        return tostr(self.proto, limit, False)

cdef public:
    cdef bool pvxs_isType(object v):
        return isinstance(v, _Type)

# sub-class hooks
Value = _Value
Type = _Type

############### client

class Cancelled(RuntimeError):
    "Cancelled from client end."
    def __init__(self, msg=None):
        RuntimeError.__init__(self, msg or "Cancelled by client")

class Disconnected(RuntimeError):
    "Channel becomes disconected."
    def __init__(self, msg=None):
        RuntimeError.__init__(self, msg or "Channel disconnected")

class Finished(Disconnected):
    "Special case of Disconnected when a Subscription has received all updates it will ever receive."
    def __init__(self, msg=None):
        Disconnected.__init__(self, msg or "Subscription Finished")

class RemoteError(RuntimeError):
    "Thrown with an error message which has been sent by a server to its remote client"

cdef public:
    cdef object _Cancelled = Cancelled
    cdef object _Disconnected = Disconnected
    cdef object _Finished = Finished
    cdef object _RemoteError = RemoteError

# can't tp_clear as we have no way to replace Operation handler (cancel?)
@cython.no_gc_clear
cdef class ClientOperation:
    cdef shared_ptr[client.Operation] op
    cdef object handler
    cdef object builder
    cdef object __weakref__

    def __init__(self, ClientProvider ctxt, basestring name,
                 handler=None, _Value value=None, builder=None, _Value pvRequest=None, get=False,
                 put=False, rpc=False):
        cdef string pvname = name.encode()
        cdef bool doGet = get
        cdef bool doPut = put
        cdef bool doRPC = rpc
        cdef client.GetBuilder bget
        cdef client.PutBuilder bput
        cdef client.RPCBuilder brpc

        if not <bool>ctxt.ctxt:
            raise RuntimeError("Context closed")

        self.handler = handler
        self.builder = builder

        if doGet and not doPut and not doRPC:
            bget = ctxt.ctxt.get(pvname)
            opHandler(bget, handler)
            if pvRequest is not None:
                bget.rawRequest(pvRequest.val)
            with nogil:
                self.op = bget.exec_()

        elif doPut and not doRPC:
            bput = ctxt.ctxt.put(pvname).fetchPresent(doGet)
            opHandler(bput, handler)
            opBuilder(bput, builder)
            if pvRequest is not None:
                bput.rawRequest(pvRequest.val)
            with nogil:
                self.op = bput.exec_()

        elif not doGet and not doPut and doRPC:
            brpc = ctxt.ctxt.rpc(pvname, value.val)
            opHandler(brpc, handler)
            if pvRequest is not None:
                brpc.rawRequest(pvRequest.val)
            with nogil:
                self.op = brpc.exec_()

        else:
            raise ValueError("Operation unsupported combination of get=, put=, and rpc=")

    def __dealloc__(self):
        self._close()

    def close(self):
        self._close()

    cdef _close(self):
        cdef shared_ptr[client.Operation] op
        cdef bool cancelled = False
        self.op.swap(op)
        if <bool>op:
            with nogil:
                cancelled = op.get().cancel()
                op.reset()
        if cancelled:
            self.handler(1, "", None)

# can't tp_clear as we have no way to replace Subscription handler (cancel?)
@cython.no_gc_clear
cdef class ClientMonitor:
    cdef shared_ptr[client.Subscription] sub
    cdef readonly object handler
    cdef object __weakref__

    def __init__(self, ClientProvider ctxt, basestring name, handler=None, _Value pvRequest=None):
        cdef string pvname = name.encode()
        cdef client.MonitorBuilder builder

        if not <bool>ctxt.ctxt:
            raise RuntimeError("Context closed")

        self.handler = handler

        builder = ctxt.ctxt.monitor(pvname) \
                      .maskConnected(True) \
                      .maskDisconnected(False)
        opEvent(builder, handler)
        if pvRequest is not None:
            builder.rawRequest(pvRequest.val)
        self.sub = builder.exec_()

    def __dealloc__(self):
        self._close()

    def close(self):
        self._close()

    cdef _close(self):
        cdef shared_ptr[client.Subscription] trash
        self.sub.swap(trash)
        if <bool>trash:
            with nogil:
                trash.get().cancel()
                trash.reset()

    def pop(self):
        cdef shared_ptr[client.Subscription] sub = self.sub # local copy to guard against concurrent _close()
        if <bool>sub:
            return monPop(sub) # will unlock/relock GIL

all_providers = WeakSet()

cdef class ClientProvider:
    def __init__(self, basestring provider, conf=None, useenv=True):
        cdef client.Config cconf
        cdef server.Config_defs_t defs

        if useenv:
            cconf.applyEnv()

        if conf is not None:
            for K,V in conf.items():
                defs[K.encode()] = V.encode()

            cconf.applyDefs(defs)

        with nogil:
            self.ctxt = cconf.build()

        all_providers.add(self)

    def __dealloc__(self):
        self.close()

    def conf(self):
        cdef client.Config_defs_t defs
        ret = {}
        if <bool>self.ctxt:
            self.ctxt.config().updateDefs(defs)
            for K,V in defs:
                ret[K.decode()] = V.decode()
        return ret

    def hurryUp(self):
        with nogil:
            self.ctxt.hurryUp()

    def close(self):
        if <bool>self.ctxt:
            all_providers.discard(self)
        with nogil:
            self.ctxt = client.Context()

    def disconnect(self, basestring name=None):
        cdef string cname
        if name is not None:
            cname = name
        if <bool>self.ctxt:
            with nogil:
                self.ctxt.cacheClear(cname, client.cacheAction.Disconnect)

    @staticmethod
    def makeRequest(basestring desc):
        cdef _Value ret = Value.__new__(Value)

        ret.val = client.Context.request().pvRequest(desc.encode()).build()

        return ret

############### server

# global provider registry
_providers = {}

all_servers = WeakSet()

cdef class Server:
    def __init__(self, conf=None, useenv=True, providers=None):
        cdef server.Config sconf
        cdef server.Config_defs_t defs
        cdef StaticProvider sprov
        cdef DynamicProvider dprov
        cdef Source src
        cdef int iorder

        if useenv:
            sconf.applyEnv()

        if conf is not None:
            for K,V in conf.items():
                defs[K.encode()] = V.encode()

            sconf.applyDefs(defs)

        with nogil:
            self.serv = sconf.build()

        for prov, order in providers:
            iorder = order
            if prov in _providers:
                prov = _providers[prov]() # lock weakref

            if isinstance(prov, StaticProvider):
                sprov = prov
                with nogil:
                    self.serv.addSource(sprov.name, sprov.src.source(), iorder)

            elif isinstance(prov, DynamicProvider):
                dprov = prov
                with nogil:
                    self.serv.addSource(dprov.name, dprov.src, iorder)

            elif isinstance(prov, Source):
                src = prov
                with nogil:
                    self.serv.addSource(src.name, src.src, iorder)

            else:
                raise ValueError("Unsupported provider type %s"%type(prov))

        all_servers.add(self)

    def __dealloc__(self):
        self.stop()
        with nogil:
            self.serv = server.Server()

    def run(self):
        if not <bool>self.serv:
            raise RuntimeError("server stop()d")
        with nogil:
            self.serv.run()

    def interrupt(self):
        if <bool>self.serv:
            with nogil:
                self.serv.interrupt()

    def start(self):
        if not <bool>self.serv:
            raise RuntimeError("server stop()d")
        with nogil:
            self.serv.start()

    def stop(self):
        if <bool>self.serv:
            all_servers.discard(self)
            with nogil:
                self.serv.stop()
        with nogil:
            self.serv = server.Server()

    def conf(self):
        cdef server.Config_defs_t defs
        ret = {}

        if not <bool>self.serv:
            raise RuntimeError("server stop()d")

        self.serv.config().updateDefs(defs)
        self.serv.clientConfig().updateDefs(defs)

        for K,V in defs:
            ret[K.decode()] = V.decode()
        return ret

    @property
    def guid(self):
        cdef server.ServerGUID guid = self.serv.config().guid
        return PyBytes_FromStringAndSize(<char*>guid.data(), guid.size())

    def tostr(self, int detail=0):
        return toString(self.serv, detail).decode()

# Avoid need for custom tp_clear to detachHandler() prior to CLEAR(handler).
@cython.no_gc_clear
cdef class SharedPV:
    cdef sharedpv.SharedPV pv
    cdef readonly object handler
    cdef object __weakref__

    def __init__(self, handler=None, options=None):
        self.pv = sharedpv.SharedPV.buildReadonly()
        if handler is not None:
            attachHandler(self.pv, handler)
        self.handler = handler

    def __dealloc__(self):
        detachHandler(self.pv)
        with nogil:
            self.pv = sharedpv.SharedPV()

    def open(self, _Value value):
        with nogil:
            self.pv.open(value.val)

    def post(self, _Value value):
        cdef data.Value temp
        with nogil:
            self.pv.post(value.val)

    def current(self):
        cdef _Value ret = Value.__new__(Value)
        with nogil:
            ret.val = self.pv.fetch()
        return ret

    def close(self, destroy=False):
        with nogil:
            self.pv.close()

    def isOpen(self):
        cdef bool ret
        with nogil:
            ret = self.pv.isOpen()
        return ret

cdef public:
    cdef sharedpv.SharedPV SharedPV_unwrap(object pv):
        cdef sharedpv.SharedPV ret
        if isinstance(pv, SharedPV):
            ret = (<SharedPV>pv).pv
        return ret

@cython.no_gc_clear
cdef class ServerOperation:
    """An in-progress Put or RPC operation from a client.
    """
    cdef shared_ptr[source.ExecOp] op
    cdef data.Value val
    cdef readonly object handler
    cdef object __weakref__

    def __dealloc__(self):
        detachCleanup(self.op)
        with nogil:
            self.op.reset()

    def pvRequest(self):
        '''pvRequest() -> Value
        Access the request Value provided by the client, which may ignored, or used to modify handling.
        '''
        cdef _Value ret = Value.__new__(Value)
        ret.val = self.op.get().pvRequest()
        return ret

    def value(self):
        '''value() -> Value
        For an RPC operation, the argument Value provided
        '''
        cdef _Value ret = Value.__new__(Value)
        ret.val = self.val
        return ret

    def name(self):
        '''name() -> str
        The PV name used by the client
        '''
        return self.op.get().name().decode()

    def peer(self):
        '''peer() -> str
        Client address
        '''
        return self.op.get().peerName().decode()

    def account(self):
        '''account() -> str
        Client identity
        '''
        return self.op.get().credentials().get().account.decode()

    def roles(self):
        '''roles() -> {str}
        Client group memberships
        '''
        ret = set()
        for role in self.op.get().credentials().get().roles():
            ret.add(role.decode())
        return ret

    def done(self, _Value value=None, basestring error=None):
        '''done(value=None, error=None)

        Signal completion of the operation. ::

          # successful completion without result (Put or RPC)
          done()
          # successful completion with result (RPC only)
          done(Value)
          # unsuccessful completion (Put or RPC)
          done(error="msg")
        '''
        cdef string msg

        if error is not None:
            msg = error.encode()
            with nogil:
                self.op.get().error(msg)

        elif value is None:
            with nogil:
                self.op.get().reply()

        else:
            with nogil:
                self.op.get().reply(value.val)

    def onCancel(self, handler):
        '''onCancel(callable|None)

        Set callable which will be invoked if the remote operation is
        cancelled by the client, or if client connection is lost.
        '''
        if handler is None:
            detachCleanup(self.op)
        else:
            attachCleanup(self.op, handler)
        self.handler = handler

    def info(self, msg):
        pass
    def warn(self, msg):
        pass

cdef public:
    cdef object ServerOperation_wrap(const shared_ptr[source.ExecOp]& op, const data.Value& val):
        cdef ServerOperation ret = ServerOperation.__new__(ServerOperation)
        ret.op = op;
        ret.val = val
        return ret

cdef class StaticProvider:
    cdef string name
    cdef sharedpv.StaticSource src
    cdef object shadow
    cdef object __weakref__

    def __init__(self, basestring name):
        self.name = name.encode()
        self.src = sharedpv.StaticSource.build()
        self.shadow = {}

    def __dealloc__(self):
        with nogil:
            self.src = sharedpv.StaticSource()

    def close(self):
        with nogil:
            self.src.close()

    def add(self, basestring name, SharedPV pv):
        cdef string cname = name.encode()
        with nogil:
            self.src.add(cname, pv.pv)
        self.shadow[name] = pv

    def remove(self, name):
        cdef string cname = name.encode()
        self.shadow.pop(name, None)
        with nogil:
            self.src.remove(cname)

    def keys(self):
        cdef cmap[string, server.SharedPV] pvs
        with nogil:
            pvs = self.src.list()
        ret = []
        for pair in pvs:
            ret.append(pair.first.decode())
        return ret

# Avoid need for custom tp_clear to disconnectDynamic() prior to CLEAR(handler).
@cython.no_gc_clear
cdef class DynamicProvider:
    cdef string name
    cdef shared_ptr[source.Source] src
    cdef readonly object handler
    cdef object __weakref__

    def __init__(self, basestring name, handler):
        self.name = name.encode()
        self.handler = handler
        self.src = createDynamic(handler)

    def __dealloc__(self):
        disconnectDynamic(self.src)

cdef class Source:
    pass




import logging
import warnings
_log = logging.getLogger(__name__)

import unittest
import sys
import random
import weakref
import gc
from unittest.case import SkipTest

from ..nt import NTScalar, NTURI
from ..server import Server, StaticProvider
from .utils import RefTestCase

import cothread

from ..client.cothread import Context, Disconnected, RemoteError
from ..server.cothread import SharedPV, _sync
from ..server import cothread as srv_cothread

class GPMHandler:
    def put(self, pv, op):
        _log.debug("putting %s <- %s", op.name(), op.value())
        cothread.Yield()  # because we can
        pv.post(op.value() * 2)
        op.done()

    def rpc(self, pv, op):
        V = op.value()
        if V.query.get('oops'):
            op.done(error='oops')
        else:
            op.done(NTScalar('i').wrap(42))

class TestGPM(RefTestCase):
    timeout = 1.0

    def _sleep(self, delay):
        cothread.Sleep(delay)

    def setUp(self):
        super(TestGPM, self).setUp()

        self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=GPMHandler())
        self.pv2 = SharedPV(handler=GPMHandler(), nt=NTScalar('d'), initial=42.0)
        self.provider = StaticProvider("serverend")
        self.provider.add('foo', self.pv)
        self.provider.add('bar', self.pv2)

    def tearDown(self):
        del self.pv
        del self.pv2
        del self.provider
        _sync()
        gc.collect()
        self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
        super(TestGPM, self).tearDown()

    def test_getput(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:
                self.assertEqual(0, C.get('foo'))

                C.put('foo', 5)

                self.assertEqual(5 * 2, C.get('foo'))

                self.assertEqual([5 * 2, 42.0], C.get(['foo', 'bar']))

                C.put(['foo', 'bar'], [6, 7])

                self.assertEqual([6 * 2, 7 * 2], C.get(['foo', 'bar']))

    def test_monitor(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                Q = cothread.EventQueue()

                with C.monitor('foo', Q.Signal, notify_disconnect=True) as sub:
                    self.assertIsInstance(Q.Wait(timeout=self.timeout), Disconnected)

                    self.assertEqual(0, Q.Wait(timeout=self.timeout))

                    C.put('foo', 2)

                    self.assertEqual(2 * 2, Q.Wait(timeout=self.timeout))

                    self.pv.close()

                    self.assertIsInstance(Q.Wait(timeout=self.timeout), Disconnected)

                    self.pv.open(3)
                    C.hurryUp()

                    self.assertEqual(3, Q.Wait(timeout=self.timeout))

class TestRPC(RefTestCase):
    timeout = 1.0

    def _sleep(self, delay):
        cothread.Sleep(delay)

    def setUp(self):
        super(TestRPC, self).setUp()

        self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=GPMHandler())
        self.provider = StaticProvider("serverend")
        self.provider.add('foo', self.pv)

    def tearDown(self):
        del self.pv
        del self.provider
        _sync()
        gc.collect()
        self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
        super(TestRPC, self).tearDown()

    def test_rpc(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                args = NTURI([
                    ('lhs', 'd'),
                    ('rhs', 'd'),
                ])

                ret = C.rpc('foo', args.wrap('foo', kws={'lhs':1, 'rhs':2}))
                _log.debug("RET %s", ret)
                self.assertEqual(ret, 42)

    def test_rpc_error(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                args = NTURI([
                    ('oops', '?'),
                ])

                with self.assertRaisesRegexp(RemoteError, 'oops'):
                    ret = C.rpc('foo', args.wrap('foo', kws={'oops':True}))

class TestFirstLast(RefTestCase):
    maxDiff = 1000
    timeout = 1.0
    mode = 'Mask'

    class Handler:
        def __init__(self):
            self.evt = cothread.Event(auto_reset=False)
            self.conn = None
        def onFirstConnect(self, pv):
            _log.debug("onFirstConnect")
            self.conn = True
            self.evt.Signal()
        def onLastDisconnect(self, pv):
            _log.debug("onLastDisconnect")
            self.conn = False
            self.evt.Signal()

    def _sleep(self, delay):
        cothread.Sleep(delay)

    def setUp(self):
        # gc.set_debug(gc.DEBUG_LEAK)
        super(TestFirstLast, self).setUp()

        self.H = self.Handler()
        self.pv = SharedPV(handler=self.H,
                        nt=NTScalar('d'),
                        options={'mapperMode':self.mode})
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)

        self.server = Server(providers=[self.sprov], isolate=True)

    def tearDown(self):
        self.server.stop()
        del self.server
        del self.sprov
        del self.pv
        del self.H
        _sync()
        gc.collect()
        self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
        super(TestFirstLast, self).tearDown()

    def testClientDisconn(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = cothread.EventQueue()
            with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                Q.Wait(timeout=self.timeout) # initial update

                _log.debug('TEST')
                self.H.evt.Wait(self.timeout) # onFirstConnect()
                self.H.evt.Reset()
                self.assertTrue(self.H.conn)

        self.H.evt.Wait(self.timeout) # onLastDisconnect()
        _log.debug('SHUTDOWN')
        self.assertFalse(self.H.conn)

    def testServerShutdown(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = cothread.EventQueue()
            with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                Q.Wait(timeout=self.timeout)

                _log.debug('TEST')
                self.H.evt.Wait(self.timeout) # initial update
                self.H.evt.Reset()
                self.assertIs(self.H.conn, True)

                self.server.stop()

                self.H.evt.Wait(self.timeout) # wait for
                _log.debug('SHUTDOWN')
                self.assertIs(self.H.conn, False)

    def testPVClose(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = cothread.EventQueue()
            with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                Q.Wait(timeout=self.timeout) # initial update

                _log.debug('TEST')
                self.H.evt.Wait(self.timeout)
                self.H.evt.Reset()
                self.assertTrue(self.H.conn)

                _log.debug('CLOSING')
                self.sprov.remove('foo') # prevent new connections while destroying
                self.pv.close(destroy=True, sync=True, timeout=self.timeout)

                _log.debug('CLOSED')
                self.assertFalse(self.H.conn)




from __future__ import print_function

import unittest
import weakref
import gc

from ..client.thread import Context, TimeoutError
from .utils import gctrace
from .utils import RefTestCase


class TestTimeout(RefTestCase):

    def setUp(self):
        super(TestTimeout, self).setUp()
        self.ctxt = Context('pva')

    def tearDown(self):
        self.ctxt.close()
        W = weakref.ref(self.ctxt)
        del self.ctxt
        gc.collect()
        C = W()
        if C is not None:
            print('trace', C)
            gctrace(C)
        self.assertIsNone(C)
        super(TestTimeout, self).tearDown()

    def test_get(self):
        R = self.ctxt.get('invalid:pv:name', timeout=0.1, throw=False)
        self.assertIsInstance(R, TimeoutError)

    def test_get_throw(self):
        self.assertRaises(TimeoutError, self.ctxt.get, 'invalid:pv:name', timeout=0.1)

    def test_put(self):
        R = self.ctxt.put('invalid:pv:name', 0, timeout=0.1, throw=False)
        self.assertIsInstance(R, TimeoutError)

    def test_put_throw(self):
        self.assertRaises(TimeoutError, self.ctxt.put, 'invalid:pv:name', 0, timeout=0.1)



#cython: language_level=2

from libcpp cimport bool
from libcpp.memory cimport shared_ptr

cdef extern from "<pvxs/sharedArray.h>" namespace "pvxs" nogil:
    cdef enum ArrayType:
        Null "ArrayType::Null"
        Bool "pvxs::ArrayType::Bool"
        Int8 "pvxs::ArrayType::Int8"
        Int16 "pvxs::ArrayType::Int16"
        Int32 "pvxs::ArrayType::Int32"
        Int64 "pvxs::ArrayType::Int64"
        UInt8 "pvxs::ArrayType::UInt8"
        UInt16 "pvxs::ArrayType::UInt16"
        UInt32 "pvxs::ArrayType::UInt32"
        UInt64 "pvxs::ArrayType::UInt64"
        Float32 "pvxs::ArrayType::Float32"
        Float64 "pvxs::ArrayType::Float64"
        String "pvxs::ArrayType::String"
        Value "pvxs::ArrayType::Value"

    size_t elementSize(ArrayType type) except+

    cdef cppclass shared_array[T]:
        shared_array()
        shared_array(const shared_array&)
        shared_array(shared_array&&)

        shared_array(size_t cnt) except+
        shared_array(T*, size_t)
        shared_array(const shared_ptr[T]&, size_t)
        # TODO omitted

        size_t size() const
        bool empty() const
        bool unique() const
        size_t clear()
        void swap(const shared_array&)

        T* data() const
        const shared_ptr[T]& dataPtr() const

        void resize(size_t cnt) except+
        void make_unique() except+

        shared_array[C] freeze[C]() except+
        shared_array[C] castTo[C]() except+
        shared_array[C] convertTo[C]() except+

        T* begin()
        T* end()

        T& operator[](size_t)
        T& at(size_t) except+

        # void only
        ArrayType original_type() const



import logging
import unittest
from functools import wraps

from qtpy.QtCore import QObject, QCoreApplication

from ..client.Qt import exceptionGuard, Disconnected, TimeoutError, Context

from ..server import Server, StaticProvider
from ..server.thread import SharedPV, _defaultWorkQueue
from ..nt import NTScalar
from .utils import RefTestCase

_log = logging.getLogger(__name__)

class waitSignal(QObject):
    def __init__(self, sig=None):
        QObject.__init__(self)
        self._T = None
        self._args = None
        self._sig = sig

    def __enter__(self):
        if self._sig is not None:
            self._sig.connect(self.slot)
        return self

    def __exit__(self,A,B,C):
        self.close()
        if self._sig is not None:
            self._sig.disconnect(self.slot)

    def close(self):
        if self._T is not None:
            self.killTimer(self._T)
            self._T = None

    @exceptionGuard
    def timerEvent(self, evt):
        _log.error('Test timeout')
        QCoreApplication.instance().quit()
        self.close()

    def slot(self, *args):
        _log.debug('signalled %s', args)
        self._args = args
        QCoreApplication.instance().quit()
        self.close()

    def wait(self, timeout=5.0):
        assert self._T is None, self._T
        self._T = self.startTimer(timeout*1000)
        try:
            self._args = None
            QCoreApplication.instance().exec_()
            if self._args is not None:
                return self._args

            else:
                raise TimeoutError()
        finally:
            self.close()

    def waitValue(self, **kws):
        E, = self.wait(**kws)
        if isinstance(E, Exception):
            raise E
        else:
            return E

class TestGPM(RefTestCase):
    def setUp(self):
        super(TestGPM, self).setUp()

        self._timedout = False
        self.app = QCoreApplication([])

        self.pv = SharedPV(nt=NTScalar('i'), initial=42)
        @self.pv.put
        def mailbox(pv, op):
            _log.debug('onPut: %s', op.value())
            pv.post(op.value())
            op.done()
        self.server = Server(providers=[{'pvname':self.pv}], isolate=True)

        self.ctxt = Context('pva', conf=self.server.conf(), useenv=False)

    def tearDown(self):
        self.ctxt.close()

        self.server.stop()
        _defaultWorkQueue.sync()
        del self.ctxt
        del self.server
        del self.pv
        del self.app
        super(TestGPM, self).tearDown()

    def test_pm(self):
        with waitSignal() as sub, waitSignal() as put:
            S = self.ctxt.monitor('pvname', sub.slot)
            E, = sub.wait()
            self.assertIsInstance(E, Disconnected)

            self.assertEqual(sub.waitValue(), 42)

            self.ctxt.put('pvname', 51, slot=put.slot)
            put.waitValue()

            self.assertEqual(sub.waitValue(), 51)



#cython: language_level=2

from libcpp cimport bool

from . cimport data

cdef extern from "<pvxs/nt.h>" namespace "pvxs::nt" nogil:
    cdef cppclass NTScalar:
        data.TypeCode value
        bool display
        bool control
        bool valueAlarm

        data.TypeDef build() except+
        data.Value create() except+



from __future__ import print_function

import logging
import sys
import time

from collections import OrderedDict

from ..wrapper import Value, Type
from .. import nt
from .utils import RefTestCase

import numpy
from numpy.testing import assert_array_almost_equal as assert_aequal

_log = logging.getLogger(__name__)

class TestScalar(RefTestCase):

    def test_float_wrap(self, code='d', value=5.0):
        NT = nt.NTScalar(code)

        V = NT.wrap(value, timestamp=None)
        self.assertEqual(V.value, value)
        self.assertEqual(V.alarm.severity, 0)
        self.assertIsNone(V.get('display'))

        NT = nt.NTScalar(code, display=True)
        V = NT.wrap({
            'value': value,
            'alarm': {
                'severity': 1,
            },
        })

        self.assertEqual(V.value, value)
        self.assertEqual(V.alarm.severity, 1)
        if code!='s':
            self.assertEqual(V.display.tolist(), [
                ('limitLow', 0.0),
                ('limitHigh', 0.0),
                ('description', u''),
                ('format', u''),
                ('units', u'')
            ])

    def test_time_wrap(self):
        NT = nt.NTScalar('d')
        V = NT.wrap(42, timestamp=None) # no timestamp
        self.assertSetEqual(V.changedSet(), {'value'})
        self.assertEqual(V['timeStamp.secondsPastEpoch'], 0)

        V = NT.wrap(42, timestamp=(1234, 5678)) # specific
        self.assertSetEqual(V.changedSet(), {
            'value',
            'timeStamp.secondsPastEpoch',
            'timeStamp.nanoseconds',
        })
        self.assertEqual(V['timeStamp.secondsPastEpoch'], 1234)
        self.assertEqual(V['timeStamp.nanoseconds'], 5678)

        V = NT.wrap(42, timestamp=1234.5) # specific
        self.assertEqual(V['timeStamp.secondsPastEpoch'], 1234)
        self.assertEqual(V['timeStamp.nanoseconds'], 500000000)

    def test_int_wrap(self):
        self.test_float_wrap(code='i', value=42)
    def test_str_wrap(self):
        self.test_float_wrap(code='s', value='foo')

    def test_float_unwrap(self, code='d', value=5.0):
        NT = nt.NTScalar(code)
        V = NT.wrap({
            'value': value,
            'alarm': {
                'severity': 1,
            },
        })

        P = nt.NTScalar.unwrap(V)
        self.assertEqual(P, value)
        self.assertEqual(P.severity, 1)

        V2 = NT.wrap(P)
        self.assertIs(V, V2)

    def test_int_unwrap(self):
        self.test_float_unwrap(code='i', value=42)
    def test_str_unwrap(self):
        self.test_float_unwrap(code='s', value='foo')

    def test_array_wrap(self):
        NT = nt.NTScalar('ad')  # array of double

        A = numpy.asarray([1.0, 5.0])
        V = NT.wrap(A)
        assert_aequal(V.value, A)
        self.assertEqual(V.alarm.severity, 0)
        self.assertTrue(V.changed('value'))

    def test_array_unwrap(self):
        NT = nt.NTScalar('ad')  # array of double
        A = numpy.asarray(range(10))[2:5]
        V = NT.wrap(A)

        P = nt.NTScalar.unwrap(V)
        self.assertIsNot(P, A)
        assert_aequal(P, A)
        self.assertEqual(P.severity, 0)

    def test_string_array_wrap(self):
        NT = nt.NTScalar('as')  # array of string

        A = ["hello", "world"]
        V = NT.wrap(A)
        self.assertEqual(V.value, A)
        self.assertEqual(V.alarm.severity, 0)
        self.assertTrue(V.changed('value'))


class TestTable(RefTestCase):

    def test_wrap(self):
        NT = nt.NTTable(columns=[
            ('a', 'i'),
            ('b', 's'),
        ])
        V = NT.wrap([
            {'a': 5, 'b': 'one'},
            {'a': 6, 'b': 'two'},
        ])

        assert_aequal(V.value.a, [5, 6])
        self.assertEqual(V.value.b, ['one', 'two'])

    def test_unwrap(self):
        T = nt.NTTable.buildType(columns=[
            ('a', 'ai'),
            ('b', 'as'),
        ])
        V = Value(T, {
            'labels': ['a', 'b'],
            'value': {
                'a': [5, 6],
                'b': ['one', 'two'],
            },
        })

        P = list(nt.NTTable.unwrap(V))

        self.assertListEqual(P, [
            OrderedDict([('a', 5), ('b', u'one')]),
            OrderedDict([('a', 6), ('b', u'two')]),
        ])


class TestURI(RefTestCase):

    def test_build(self):
        NT = nt.NTURI([
            ('a', 'I'),
            ('b', 's'),
            ('c', ('S', None, [
                ('x', 'd'),
                ('y', 'd'),
            ])),
        ])

        V = NT.wrap('fn', (5,))
        self.assertEqual(V.query.a, 5)
        self.assertRaises(AttributeError, lambda: V.query.b)
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (6, 'foo'))
        self.assertEqual(V.query.a, 6)
        self.assertEqual(V.query.b, 'foo')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (7,), {'b': 'bar'})
        self.assertEqual(V.query.a, 7)
        self.assertEqual(V.query.b, 'bar')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (), {'a': 8, 'b': 'bar'})
        self.assertEqual(V.query.a, 8)
        self.assertEqual(V.query.b, 'bar')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (), {'a': 8, 'b': 'bar', 'c': {'x': 1, 'y': 2}})
        self.assertEqual(V.query.a, 8)
        self.assertEqual(V.query.c.x, 1)
        self.assertEqual(V.query.c.y, 2)


class TestEnum(RefTestCase):

    def testStore(self):
        T = nt.NTEnum.buildType()

        V = Value(T, {
            'value.choices': ['zero', 'one', 'two'],
        })

        V.value = 'one'

        self.assertEqual(V.value.index, 1)

        V.value = '2'

        self.assertEqual(V.value.index, 2)

        V.value = 1

        self.assertEqual(V.value.index, 1)

    def testStoreBad(self):
        V = Value(nt.NTEnum.buildType(), {
            'value.choices': ['zero', 'one', 'two'],
        })

        V.value.index = 42

        def fn():
            V.value = self
        self.assertRaises(TypeError, fn)

        self.assertEqual(V.value.index, 42)

        def fn():
            V.value = 'other'
        self.assertRaises(ValueError, fn)

        self.assertEqual(V.value.index, 42)

        if sys.version_info >= (3, 0):
            V.value.choices = []

            def fn():
                V.value = '1'
            self.assertWarns(UserWarning, fn)  # warns of empty choices

            self.assertEqual(V.value.index, 1)

    def testSubStore(self):
        V = Value(Type([
            ('a', nt.NTEnum.buildType()),
            ('b', nt.NTEnum.buildType()),
        ]), {
            'a.value.choices': ['A', 'B'],
            'b.value.choices': ['X', 'Y'],
        })

        V.a = {'value': 'B'}

        self.assertEqual(V.a.value.index, 1)
        self.assertEqual(V.b.value.index, 0)

    def testWrap(self):
        W = nt.NTEnum()
        V = W.wrap({'index':1, 'choices':['X','Y']})

        self.assertEqual(V.value.index, 1)
        self.assertEqual(V.value.choices, ['X','Y'])

        W = nt.NTEnum()
        V = W.wrap(0)

        self.assertEqual(V.value.index, 0)
        self.assertEqual(V.value.choices, [])

    def testAssign(self):
        W = nt.NTEnum()
        V = nt.NTEnum.buildType()()
        W.assign(V, 1)

        self.assertEqual(V.value.index, 1)
        self.assertEqual(V.value.choices, [])

        V.value.choices = ['A', 'B']
        W.assign(V, 'A')

        self.assertEqual(V.value.index, 0)
        self.assertEqual(V.value.choices, ['A', 'B'])

    def testUnwrap(self):
        W = nt.NTEnum()
        V = nt.NTEnum.buildType()()

        U = W.unwrap(V)
        self.assertEqual(U, 0)
        self.assertIsNone(U.choice)

        V.value.index = 1
        V.value.choices = ['A', 'B']
        U = W.unwrap(V)
        self.assertEqual(U, 1)
        self.assertEqual(U.choice, 'B')
        self.assertEqual(W._choices, ['A', 'B'])

class TestArray(RefTestCase):

    def test_unwrap_None(self):
        V = Value(nt.NTNDArray.buildType(), {})

        img = nt.NTNDArray.unwrap(V)

        self.assertIsInstance(img, numpy.ndarray)
        assert_aequal(img.shape, (0,))

    def test_zero_length(self):
        V = Value(nt.NTNDArray.buildType(), {
            'value': numpy.arange(0),
            'dimension': [
                {'size': 3}, # X, columns
                {'size': 0}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 0},
            ],
        })

        img = nt.NTNDArray.unwrap(V)

        self.assertIsInstance(img, numpy.ndarray)
        assert_aequal(img.shape, (0,3))

    def test_unwrap_mono(self):
        pixels = numpy.asarray([  # 2x3
            [0, 1, 2],
            [3, 4, 5],
        ])
        # check my understanding of numpy
        self.assertTupleEqual(pixels.shape, (2, 3)) # inner-most right (in a pixel loop)
        assert_aequal(pixels.flatten(), [0, 1, 2, 3, 4, 5]) # row major

        V = Value(nt.NTNDArray.buildType(), {
            'value': numpy.arange(6),
            'dimension': [
                {'size': 3}, # X, columns
                {'size': 2}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 0},
            ],
        })

        img = nt.NTNDArray.unwrap(V)

        self.assertEqual(img.shape, (2, 3))
        assert_aequal(img, pixels)

        V2 = nt.NTNDArray().wrap(img)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size, msg=V.dimension)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size, msg=V.dimension)

    def test_unwrap_3d(self):
        # scipy.misc.face().shape==(768, 1024, 3)
        # in AD world this is [3, 1024, 768) w/ RGB1

        # we test with 4x2 RGB1
        pixels = numpy.array([
            [(1,0,0), (2,0,0), (3,0,0), (4,0,0)],
            [(5,0,0), (6,0,0), (7,0,0), (8,0,0)],
        ], dtype='u1')
        self.assertEqual(pixels.shape, (2, 4, 3))

        # manually construct matching NTNDArray
        V = Value(nt.NTNDArray.buildType(), {
            'value': ('ubyteValue', pixels.flatten()),
            'dimension': [
                {'size': 3}, # "color"
                {'size': 4}, # X, columns
                {'size': 2}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 2},
            ],
        })
        self.assertEqual(V.value.dtype, pixels.dtype)

        img = nt.NTNDArray.unwrap(V)

        self.assertEqual(img.shape, (2, 4, 3))
        assert_aequal(img, pixels)
        self.assertEqual(img.dtype, pixels.dtype)
        self.assertDictEqual(img.attrib, {u'ColorMode':2}) # RGB1

        # round trip
        V2 = nt.NTNDArray().wrap(img)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size)
        self.assertEqual(V.dimension[2].size, V2.dimension[2].size)
        self.assertEqual(V2.value.dtype, pixels.dtype)
        self.assertEqual(V2.attribute[0].name, u'ColorMode')
        self.assertEqual(V2.attribute[0].value, 2)

        # wrap up raw array  (no pixels.attrib)
        V2 = nt.NTNDArray().wrap(pixels)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size)
        self.assertEqual(V.dimension[2].size, V2.dimension[2].size)
        self.assertEqual(V2.value.dtype, pixels.dtype)
        self.assertEqual(V2.attribute[0].name, u'ColorMode')
        self.assertEqual(V2.attribute[0].value, 2)



#cython: language_level=2

from libcpp cimport bool

cdef extern from "<pvxs/version.h>" namespace "pvxs" nogil:
    enum: PVXS_VERSION
    enum: EPICS_VERSION_INT

    const char *version_str()
    unsigned long version_int()
    unsigned long version_abi_int()
    bool version_abi_check()



from __future__ import print_function

import sys
import os
import logging
import time
import socket
import threading
import pprint
import json
import re
import sqlite3

from functools import wraps, reduce

from .nt import NTScalar, Type, NTTable
from .server import Server, StaticProvider, removeProvider
from .server.thread import SharedPV, RemoteError
from .client.thread import Context
from . import set_debug, listRefs
from . import _gw
from .asLib import Engine, ACFError
from .asLib.pvlist import PVList
from .test.utils import RegularNamedTemporaryFile as NamedTemporaryFile

if sys.version_info >= (3, 0):
    unicode = str

_log = logging.getLogger(__name__)
_log_audit = logging.getLogger(__name__+'.audit')

def uricall(fn):
    @wraps(fn)
    def rpc(self, pv, op):
        ret = fn(self, op, **dict(op.value().query.items()))
        op.done(ret)
    return rpc

class TestChannel(object):
    def __init__(self, name):
        self.name = name # logging only
        self.perm = None
    def access(self, **kws):
        self.perm = kws

class TableBuilder(object):
    def __init__(self, colinfo):
        self.labels, cols = [], []
        for type, name, label in colinfo:
            self.labels.append(label)
            cols.append((name, 'a'+type))

        self.type = NTTable.buildType(cols)

    def wrap(self, values):
        S, NS = divmod(time.time(), 1.0)
        ret = self.type({
            'labels':self.labels,
            'timeStamp': {
                'secondsPastEpoch': S,
                'nanoseconds': NS * 1e9,
            },
        })
        # unzip list of tuple into tuple of lists
        cols = list([] for k in ret.value)

        for row in values:
            for I, C in zip(row, cols):
                C.append(I)

        for k, C in zip(ret.value, cols):
            ret.value[k] = C

        return ret

    def unwrap(self, value):
        return value

class RefAdapter(object):
    def __init__(self):
        self.type = NTTable.buildType([('type', 'as'), ('count', 'aI'), ('delta', 'ai')])
        self.prev = {}
        self._labels = ['Type', 'Count', 'Delta']

    def wrap(self, cnts):
        kcur = set(cnts)
        kprev = set(self.prev)

        update = []

        added = kcur - kprev
        for k in added:
            update.append((k, cnts[k], cnts[k]))

        removed = kprev - kcur
        for k in removed:
            update.append((k, 0, -self.prev[k]))

        for k in kcur&kprev:
            c, p = cnts[k], self.prev[k]
            if c!=p:
                update.append((k, c, c-p))

        update.sort(key=lambda t:t[0])
        self.prev = cnts

        Ns, Cs, Ds = [], [], []
        for N, C, D in update:
            Ns.append(N)
            Cs.append(C)
            Ds.append(D)

        V = self.type({
            'value.type':Ns,
            'value.count':Cs,
            'value.delta':Ds,
            'timeStamp.secondsPastEpoch':time.time(),
        })
        if self._labels is not None:
            self._labels, V.labels = None, self._labels

        return V

    def unwrap(self, V):
        return V

statsType = Type([
    ('ccacheSize', NTScalar.buildType('L')),
    ('mcacheSize', NTScalar.buildType('L')),
    ('gcacheSize', NTScalar.buildType('L')),
    ('banHostSize', NTScalar.buildType('L')),
    ('banPVSize', NTScalar.buildType('L')),
    ('banHostPVSize', NTScalar.buildType('L')),
], id='epics:p2p/Stats:1.0')

permissionsType = Type([
    ('pv', 's'),
    ('account', 's'),
    ('peer', 's'),
    ('roles', 'as'),
    ('asg', 's'),
    ('asl', 'i'),
    ('permission', ('S', None, [
        ('put', '?'),
        ('rpc', '?'),
        ('uncached', '?'),
        ('audit', '?'),
    ])),
], id='epics:p2p/Permission:1.0')

asDebugType = NTTable.buildType([
    ('asg', 'as'),
    ('var', 'as'),
    ('value', 'ad'),
    ('connected', 'a?'),
])

class GWStats(object):
    """I manage statistics for all GWHandler instances
    """
    def __init__(self, statsdb=None):
        self.statsdb = statsdb

        self.handlers = [] # GWHandler instances
        self.servers = [] # _p4p.Server instances

        self._pvs = {} # name suffix -> SharedPV

        if not statsdb:
            self.__tempdb = NamedTemporaryFile()
            self.statsdb = self.__tempdb.name
            _log.debug("Using temporary stats db: %s", self.statsdb)

        with sqlite3.connect(self.statsdb) as C:
            C.executescript("""
                DROP TABLE IF EXISTS us;
                DROP TABLE IF EXISTS usbyname;
                DROP TABLE IF EXISTS usbypeer;
                DROP TABLE IF EXISTS ds;
                DROP TABLE IF EXISTS dsbyname;
                DROP TABLE IF EXISTS dsbypeer;

                CREATE TABLE us(
                    usname REAL NOT NULL,
                    optx INTEGER NOT NULL,
                    oprx REAL NOT NULL,
                    peer STRING NOT NULL,
                    trtx REAL NOT NULL,
                    trrx REAL NOT NULL
                );
                CREATE TABLE ds(
                    usname STRING NOT NULL,
                    dsname STRING NOT NULL,
                    optx REAL NOT NULL,
                    oprx REAL NOT NULL,
                    account STRING NOT NULL,
                    peer STRING NOT NULL,
                    trtx REAL NOT NULL,
                    trrx REAL NOT NULL
                );

                CREATE TABLE usbyname(
                    usname STRING NOT NULL,
                    tx REAL NOT NULL,
                    rx REAL NOT NULL
                );
                CREATE TABLE dsbyname(
                    usname STRING NOT NULL,
                    tx REAL NOT NULL,
                    rx REAL NOT NULL
                );

                CREATE TABLE usbypeer(
                    peer STRING NOT NULL,
                    tx REAL NOT NULL,
                    rx REAL NOT NULL
                );
                CREATE TABLE dsbypeer(
                    account STRING NOT NULL,
                    peer STRING NOT NULL,
                    tx REAL NOT NULL,
                    rx REAL NOT NULL
                );
            """)

        self._pvs['clients'] = self.clientsPV = SharedPV(nt=NTScalar('as'), initial=[])

        self._pvs['cache'] = self.cachePV = SharedPV(nt=NTScalar('as'), initial=[])

        self._pvs['stats'] = self.statsPV = SharedPV(initial=statsType())

        self._pvs['poke'] = self.pokeStats = SharedPV(nt=NTScalar('i'), initial=0)
        @self.pokeStats.put
        def pokeStats(pv, op):
            self.update_stats()
            op.done()

        self._pvs['refs'] = self.refsPV = SharedPV(nt=RefAdapter(), initial={})

        self._pvs['StatsTime'] = self.statsTime = SharedPV(nt=NTScalar('d'), initial=0.0)

        # faulthandler.dump_traceback() added w/ py 3.3
        # however, file= arg. added w/ 3.5
        if sys.version_info>=(3,5):
            self._pvs['threads'] = stackstrace = SharedPV(nt=NTScalar('s'), initial='RPC only')
            @stackstrace.rpc
            def showStacks(pv, op):
                import faulthandler
                from tempfile import TemporaryFile
                with TemporaryFile('r+') as F:
                    faulthandler.dump_traceback(file=F)
                    F.seek(0)
                    V = pv.current().raw
                    V.unmark()
                    V['value'] = F.read()
                    op.done(V)

        # PVs for bandwidth usage statistics.
        # 2x tables: us, ds
        # 2x groupings: by PV and by peer
        # 2x directions: Tx and Rx

        def addpv(dir='TX', suffix=''):
            pv = SharedPV(nt=TableBuilder([
                ('s', 'name', 'PV'),
                ('d', 'rate', dir+' (B/s)'),
            ]), initial=[])
            self._pvs[suffix] = pv
            return pv

        self.tbl_usbypvtx = addpv(dir='TX', suffix='us:bypv:tx')
        self.tbl_usbypvrx = addpv(dir='RX', suffix='us:bypv:rx')

        self.tbl_dsbypvtx = addpv(dir='TX', suffix='ds:bypv:tx')
        self.tbl_dsbypvrx = addpv(dir='RX', suffix='ds:bypv:rx')

        def addpv(dir='TX', suffix=''):
            pv = SharedPV(nt=TableBuilder([
                ('s', 'name', 'Server'),
                ('d', 'rate', dir+' (B/s)'),
            ]), initial=[])
            self._pvs[suffix] = pv
            return pv

        self.tbl_usbyhosttx = addpv(dir='TX', suffix='us:byhost:tx')
        self.tbl_usbyhostrx = addpv(dir='RX', suffix='us:byhost:rx')

        def addpv(dir='TX', suffix=''):
            pv = SharedPV(nt=TableBuilder([
                ('s', 'account', 'Account'),
                ('s', 'name', 'Client'),
                ('d', 'rate', dir+' (B/s)'),
            ]), initial=[])
            self._pvs[suffix] = pv
            return pv

        self.tbl_dsbyhosttx = addpv(dir='TX', suffix='ds:byhost:tx')
        self.tbl_dsbyhostrx = addpv(dir='RX', suffix='ds:byhost:rx')

    def bindto(self, provider, prefix):
        'Add myself to a StaticProvider'

        for suffix, pv in self._pvs.items():
            provider.add(prefix+suffix, pv)

    def sweep(self):
        for handler in self.handlers:
            handler.sweep()

    def update_stats(self, norm):
        T0 = time.time()
        self.refsPV.post(listRefs())

        with sqlite3.connect(self.statsdb) as C:
            C.executescript('''
                DELETE FROM us;
                DELETE FROM ds;
                DELETE FROM usbyname;
                DELETE FROM dsbyname;
                DELETE FROM usbypeer;
                DELETE FROM dsbypeer;
            ''')

            for handler in self.handlers:
                C.executemany('INSERT INTO us VALUES (?,?,?,?,?,?)', handler.provider.report(norm))

            for server in self.servers:
                C.executemany('INSERT INTO ds VALUES (?,?,?,?,?,?,?,?)', _gw.Server_report(server, norm))

            C.executescript('''
                INSERT INTO usbyname SELECT usname, sum(optx), sum(oprx) FROM us GROUP BY usname;
                INSERT INTO dsbyname SELECT usname, sum(optx), sum(oprx) FROM ds GROUP BY usname;
                INSERT INTO usbypeer SELECT peer, max(trtx), max(trrx) FROM us GROUP BY peer;
                INSERT INTO dsbypeer SELECT account, peer, max(trtx), max(trrx) FROM ds GROUP BY peer;
            ''')

            #TODO: create some indicies to speed up these queries?

            self.tbl_usbypvtx.post(C.execute('SELECT usname, tx as rate FROM usbyname ORDER BY rate DESC LIMIT 10'))
            self.tbl_usbypvrx.post(C.execute('SELECT usname, rx as rate FROM usbyname ORDER BY rate DESC LIMIT 10'))

            self.tbl_usbyhosttx.post(C.execute('SELECT peer, tx as rate FROM usbypeer ORDER BY rate DESC LIMIT 10'))
            self.tbl_usbyhostrx.post(C.execute('SELECT peer, rx as rate FROM usbypeer ORDER BY rate DESC LIMIT 10'))

            self.tbl_dsbypvtx.post(C.execute('SELECT usname, tx as rate FROM dsbyname ORDER BY rate DESC LIMIT 10'))
            self.tbl_dsbypvrx.post(C.execute('SELECT usname, rx as rate FROM dsbyname ORDER BY rate DESC LIMIT 10'))

            self.tbl_dsbyhosttx.post(C.execute('SELECT account, peer, tx as rate FROM dsbypeer ORDER BY rate DESC LIMIT 10'))
            self.tbl_dsbyhostrx.post(C.execute('SELECT account, peer, rx as rate FROM dsbypeer ORDER BY rate DESC LIMIT 10'))

            self.clientsPV.post([row[0] for row in C.execute('SELECT DISTINCT peer FROM us')])

        statsSum = {'ccacheSize.value':0, 'mcacheSize.value':0, 'gcacheSize.value':0,
                    'banHostSize.value':0, 'banPVSize.value':0, 'banHostPVSize.value':0}
        stats = [handler.provider.stats() for handler in self.handlers]
        for key in statsSum:
            for stat in stats:
                statsSum[key] += stat[key]
        self.statsPV.post(statsType(statsSum))

        cachepvs = list(reduce(set.__or__, [handler.provider.cachePeek() for handler in self.handlers], set()))
        cachepvs.sort()
        self.cachePV.post(cachepvs)

        T1 = time.time()

        self.statsTime.post(T1-T0)

class GWHandler(object):
    def __init__(self, acf, pvlist, readOnly=False):
        self.acf, self.pvlist = acf, pvlist
        self.readOnly = readOnly
        self.channels_lock = threading.Lock()
        self.channels = {}

        self.provider = None
        self.getholdoff = None


    def testChannel(self, pvname, peer):
        _log.debug('%s Searching for %s', peer, pvname)
        usname, _asg, _asl = self.pvlist.compute(pvname, peer.split(':',1)[0])

        if not usname:
            _log.debug("Not allowed: %s by %s", pvname, peer)
            return self.provider.BanHostPV
        else:
            ret = self.provider.testChannel(usname.encode('UTF-8'))
            _log.debug("allowed: %s by %s -> %s", pvname, peer, ret)
            return ret

    def makeChannel(self, op):
        _log.debug("Create %s by %s", op.name, op.peer)
        peer = op.peer.split(':',1)[0]
        usname, asg, asl = self.pvlist.compute(op.name, peer)
        if not usname or self.provider.testChannel(usname.encode())!=self.provider.Claim:
            return None
        chan = op.create(usname.encode('UTF-8'))

        with self.channels_lock:
            try:
                channels = self.channels[op.peer]
            except KeyError:
                self.channels[op.peer] = channels = []
            channels.append(chan)

        try:
            if not self.readOnly: # default is RO
                self.acf.create(chan, asg, op.account, peer, asl, op.roles)
            if self.getholdoff is not None:
                chan.access(holdoff=self.getholdoff)
        except:
            # create() should fail secure.  So allow this client to
            # connect R/O.  We already acknowledged the search, so
            # if we fail here the client will go into a reset loop.
            _log.exception("Default restrictive for %s from %s", op.name, op.peer)
        return chan

    def audit(self, msgs):
        for msg in msgs:
            _log_audit.info('%s', msg)

    def sweep(self):
        self.provider.sweep()
        replace = {}
        with self.channels_lock:
            for K, chans in self.channels.items():
                chans = [chan for chan in chans if not chan.expired]
                if chans:
                    replace[K] = chans
            self.channels = replace

    @uricall
    def asTest(self, op, pv=None, user=None, peer=None, roles=[]):
        if not user:
            user = op.account()
            if not roles:
                roles = op.roles()
        peer = peer or op.peer().split(':')[0]
        _log.debug("asTest %s %s %s", user, roles, peer)

        if not user or not peer or not pv:
            raise RemoteError("Missing required arguments pv= user= and peer=")

        peer = socket.gethostbyname(peer)
        pvname, asg, asl = self.pvlist.compute(pv.encode('UTF-8'), peer)
        if not pvname:
            raise RemoteError("Denied")

        chan=TestChannel('<asTest>')
        self.acf.create(chan, asg, user, peer, asl, roles)

        return permissionsType({
            'pv':pv,
            'account':user,
            'peer':peer,
            'roles':roles,
            'asg':asg,
            'asl':asl,
            'permission':chan.perm,
        })

    @uricall
    def asDebug(self, op):
        return asDebugType(self.acf.report())

def readnproc(args, fname, fn, **kws):
    try:
        if fname:
            fullname = os.path.join(os.path.dirname(args.config), fname)
            args._all_config_files.append(fullname)
            with open(fullname, 'r') as F:
                data = F.read()
        else:
            data = ''
        return fn(data, **kws)
    except IOError as e:
        _log.error('In "%s" : %s', fname, e)
        sys.exit(1)
    except RuntimeError as e:
        _log.error('In "%s" : %s', fname, e)
        sys.exit(1)
    except ACFError as e:
        _log.error('In "%s" %s', fname, e)
        sys.exit(1)

def comment_sub(M):
    '''Replace C style comment with equivalent whitespace, includeing newlines,
       to preserve line and columns numbers in parser errors (py3 anyway)
    '''
    return re.sub(r'[^\n]', ' ', M.group(0))

def jload(raw):
    '''Parse JSON including C style comments
    '''
    return json.loads(re.sub(r'/\*.*?\*/', comment_sub, raw, flags=re.DOTALL))

def getargs():
    from argparse import ArgumentParser
    P = ArgumentParser()
    P.add_argument('config', help='Config file')
    P.add_argument('--no-ban-local', action='store_true',
                    help='Legacy option.  Ignored')
    P.add_argument('-v', '--verbose', action='store_const', const=logging.DEBUG, default=logging.INFO,
                   help='Enable basic logging with DEBUG level')
    P.add_argument('--logging', help='Use logging config from file (JSON in dictConfig format)')
    P.add_argument('--debug', action='store_true',
                   help='Enable extremely verbose low level PVA debugging')
    P.add_argument('-T', '--test-config', action='store_true',
                   help='Read and validate configuration files, then exit w/o starting a gateway.'+
                   '  Also prints the names of all configuration files read.')
    P.add_argument('--example-config', action='store_true',
                   help='Write an example configuration file and exit.  "--example-config -" writes to stdout')
    P.add_argument('--example-systemd', action='store_true',
                   help='Write an example systemd unit file and exit  "--example-systemd -" writes to stdout')
    return P

class App(object):

    def __init__(self, args):
        args._all_config_files = [args.config]
        with open(args.config, 'r') as F:
            jconf = F.read()
        try:
            # we substitute comments with whitespace to keep correct line and column numbers
            # in error messages.
            jconf = jload(jconf)
            jver = jconf.get('version', 0)
            if jver not in (1,2):
                _log.error('Warning: config file version %d not in range [1, 2]\n'%jver)
        except ValueError as e:
            _log.error('Syntax Error in %s: %s\n'%(args.config, e.args))
            sys.exit(1)

        if not args.test_config:
            self.stats = GWStats(jconf.get('statsdb'))

        clients = {}
        statusprefix = None

        names = [jcli['name'] for jcli in jconf['clients']]
        if len(names)!=len(set(names)):
            _log.error('Duplicate client names: %s', names)
        del names

        for jcli in jconf['clients']:
            name = jcli['name']
            client_conf = {
                'EPICS_PVA_ADDR_LIST':jcli.get('addrlist',''),
                'EPICS_PVA_AUTO_ADDR_LIST':{True:'YES', False:'NO'}[jcli.get('autoaddrlist',True)],
            }
            if 'bcastport' in jcli:
                client_conf['EPICS_PVA_BROADCAST_PORT'] = str(jcli['bcastport'])
            if 'serverport' in jcli:
                client_conf['EPICS_PVA_SERVER_PORT'] = str(jcli['serverport'])

            _log.info("Client %s input config:\n%s", name, pprint.pformat(client_conf))
            if args.test_config:
                clients[name] = None
            else:
                clients[name] = Context(jcli.get('provider', u'pva'), client_conf)

        servers = self.servers = {}

        # pre-process 'servers' to expand 'interface' list
        new_servers = []
        for jsrv in jconf['servers']:
            iface = jsrv.get('interface') or ['0.0.0.0']

            if jver==1:
                # version 1 only allowed one interface.
                #  'interface':'1.2.3.4'
                # version 2 allows a list
                #  'interface':['1.2.3.4']
                if isinstance(iface, list):
                    _log.warning('Server interface list should specify JSON scheme version 2')
                else:
                    # be forgiving
                    iface = [iface]

            if len(jsrv.get('addrlist',''))>0 and len(iface)>1:
                _log.warning('Server entries for more than one interface must not specify addrlist.')
                _log.warning('Each server interface will attempt to send beacons to all destinations')
                jsrv.pop('addrlist')

            base_name = jsrv['name']
            for idx, iface in enumerate(iface):
                jsrv = jsrv.copy()
                jsrv['name'] = '%s_%d'%(base_name, idx)
                jsrv['interface'] = iface

                new_servers.append(jsrv)

        jconf['servers'] = new_servers
        del new_servers

        names = [jsrv['name'] for jsrv in jconf['servers']]
        if len(names)!=len(set(names)):
            _log.error('Duplicate server names: %s', names)
        del names

        # various objects which shouldn't be GC'd until server shutdown,
        # but aren't otherwise accessed after startup.
        self.__lifesupport = []

        gwclients = []

        for jsrv in jconf['servers']:
            name = jsrv['name']

            providers = []

            server_conf = {
                'EPICS_PVAS_INTF_ADDR_LIST':jsrv.get('interface', '0.0.0.0'),
                'EPICS_PVAS_BEACON_ADDR_LIST':jsrv.get('addrlist', ''),
                'EPICS_PVAS_AUTO_BEACON_ADDR_LIST':{True:'YES', False:'NO'}[jsrv.get('autoaddrlist',True)],
                'EPICS_PVAS_IGNORE_ADDR_LIST':jsrv.get('ignoreaddr', ''),
            }
            if 'bcastport' in jsrv:
                server_conf['EPICS_PVAS_BROADCAST_PORT'] = str(jsrv['bcastport'])
            if 'serverport' in jsrv:
                server_conf['EPICS_PVAS_SERVER_PORT'] = str(jsrv['serverport'])

            # pick client to use for ACF INP*
            aclient = jsrv.get('acf_client')
            if aclient is None:
                if len(jsrv['clients'])>1 and 'access' in jsrv:
                    _log.warning('Multiple clients and ACF is ambigious.  Add key \'acf_client\' to disambiguate')
                if len(jsrv['clients'])>0:
                    aclient = jsrv['clients'][0]

            ctxt = None
            if aclient is not None and not args.test_config:
                ctxt = clients[aclient]

            access = readnproc(args, jsrv.get('access', ''), Engine, ctxt=ctxt)
            pvlist = readnproc(args, jsrv.get('pvlist', ''), PVList)

            if args.test_config:
                continue

            statusp = StaticProvider(u'gwsts.'+name)
            providers = [statusp]
            self.__lifesupport += [statusp]

            try:
                for client in jsrv['clients']:
                    pname = u'gws.%s.%s'%(name, client)

                    client = clients[client]

                    handler = GWHandler(access, pvlist, readOnly=jconf.get('readOnly', False))
                    handler.getholdoff = jsrv.get('getholdoff')

                    if not args.test_config:
                        handler.provider = _gw.Provider(pname, client, handler) # implied installProvider()
                        providers.append((handler.provider, 10))

                    self.__lifesupport += [client]
                    gwclients +=[handler.provider]
                    self.stats.handlers.append(handler)

                if 'statusprefix' in jsrv:
                    self.stats.bindto(statusp, jsrv['statusprefix'])

                    handler.asTestPV = SharedPV(nt=NTScalar('s'), initial="Only RPC supported.")
                    handler.asTestPV.rpc(handler.asTest) # TODO this is a deceptive way to assign
                    statusp.add(jsrv['statusprefix']+'asTest', handler.asTestPV)

                    handler.asDebugPV = SharedPV(nt=NTScalar('s'), initial="Only RPC supported.")
                    handler.asDebugPV.rpc(handler.asDebug) # TODO this is a deceptive way to assign
                    statusp.add(jsrv['statusprefix']+'asDebug', handler.asDebugPV)

                    # prevent client from searching for our status PVs
                    for spv in statusp.keys():
                        handler.provider.forceBan(usname=spv.encode('utf-8'))

                try:
                    server = Server(providers=providers,
                                    conf=server_conf, useenv=False)
                except RuntimeError:
                    _log.exception("Unable to create server %s", pprint.pformat(server_conf))
                    sys.exit(1)
                self.stats.servers.append(server._S)
                # we're live now...

                _log.info("Server effective config %s :\n%s", name, pprint.pformat(server.conf()))

                for spv in statusp.keys():
                    _log.info('Status PV: %s', spv)

            finally:
                [removeProvider(pname) for pname in providers[1:]]

            # keep it all from being GC'd
            servers[name] = server

        if args.test_config:
            return

        # inform GW clients of GW server GUIDs to be ignored to prevent loops
        _log.info('Setup GW clients to ignore GW servers')
        gwservers = [serv._S for serv in servers.values()]
        for gwclient in gwclients:
            gwclient.ignoreByGUID(gwservers)

        # try to prevent client -> server loops.
        # servers and clients already running, so possible race...

        if args.no_ban_local:
            _log.info('--no-ban-local is no longer needed, and is a no-op')


    def run(self):
        # needs to be longer than twice the longest search interval
        period = 30
        try:
            while True:
                # periodic cleanup of channel cache
                _log.debug("Channel Cache sweep")
                try:
                    self.stats.sweep()
                    self.stats.update_stats(period)
                except:
                    _log.exception("Error during periodic sweep")

                self.sleep(period)
        except KeyboardInterrupt:
            pass
        finally:
            [server.stop() for server in self.servers.values()]

    @staticmethod
    def sleep(dly):
        time.sleep(dly)

def main(args=None):
    args = getargs().parse_args(args)

    catfile = None
    if args.example_config:
        catfile = 'example.conf'
    elif args.example_systemd:
        catfile = 'pvagw@.service'

    if catfile is not None:
        if args.config=='-':
            O = sys.stdout
            I = '%i'
            conf = '/etc/pvagw/%i.conf'
            print('# eg. save as /etc/systemd/system/pvagw@.service', file=sys.stderr)
        else:
            O = open(args.config, 'w')
            I = os.path.splitext(os.path.basename(args.config))[0]
            conf = os.path.abspath(args.config)

        pythonpath=os.environ.get('PYTHONPATH','').split(os.pathsep)
        modroot = os.path.dirname(os.path.dirname(__file__)) # directory containing p4p/gw.py
        if modroot not in sys.path:
            pythonpath = [modroot]+pythonpath

        with open(os.path.join(os.path.dirname(__file__), catfile), 'r') as F:
            O.write(F.read()%{
                'python':sys.executable,
                'inst':I,
                'conf':conf,
                'pythonpath':os.pathsep.join(pythonpath),
            })

        O.close()
        sys.exit(0)

    if args.logging is not None:
        with open(args.logging, 'r') as F:
            jconf = F.read()
        try:
            from logging.config import dictConfig
            dictConfig(jload(jconf))
        except ValueError as e:
            sys.stderr.write('%s Logging config Error: %s\n'%(args.logging, e.args))
            sys.exit(1)

    else:
        logging.basicConfig(level=args.verbose)
    if args.debug:
        set_debug(logging.DEBUG)

    app = App(args)
    if args.test_config:
        _log.info('Configuration valid')
        for fname in args._all_config_files:
            print(fname)
    else:
        app.run()

    return 0

if __name__=='__main__':
    main()



#cython: language_level=2

from libcpp.memory cimport shared_ptr
from libcpp.string cimport string

from pvxs cimport client
from pvxs cimport source
from pvxs cimport server

cdef class ClientProvider:
    cdef client.Context ctxt

cdef class Source:
    cdef string name
    cdef shared_ptr[source.Source] src


cdef class Server:
    cdef server.Server serv
    cdef object __weakref__



#cython: language_level=2

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.map cimport map
from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
from libcpp.utility cimport pair

from . cimport client
from .util cimport ServerGUID
from .source cimport Source
from .sharedpv cimport SharedPV

cdef extern from "<pvxs/server.h>" namespace "pvxs::server" nogil:
    ctypedef map[string, string] Config_defs_t "pvxs::server::Config::defs_t"

    ctypedef client.Report Report

    cdef cppclass Config:
        vector[string] interfaces
        vector[string] beaconDestinations
        unsigned short tcp_port
        unsigned short udp_port
        bool auto_beacon
        ServerGUID guid

        @staticmethod
        Config fromEnv() except+
        @staticmethod
        Config fromDefs(const map[string, string]&) except+
        @staticmethod
        Config isolated() except+

        Config& applyEnv() except+
        Config& applyDefs(const Config_defs_t&) except+
        void updateDefs(Config_defs_t& defs) except+

        void expand() except+
        Server build() except+

    cdef cppclass Server:
        Server()
        Server(const Config&) except+

        Server& start() except+
        Server& stop() except+
        Server& run() except+
        Server& interrupt() except+

        const Config& config() except+
        client.Config clientConfig() except+

        Server& addPV(const string& name, const SharedPV& pv) except+
        Server& removePV(const string& name) except+

        Server& addSource(const string& name, const shared_ptr[Source]& src, int order) except+
        shared_ptr[Source] removeSource(const string& name, int order) except+

        shared_ptr[Source] getSource(const string& name, int order) except+

        vector[pair[string, int]] listSource() except+

        Report report() except+



import sys
import unittest

if sys.version_info<(3,6):
    class TestDummy(unittest.TestCase):
        def test_asyncio(self):
            raise unittest.SkipTest("asyncio needs Py >=3.6")

else:
    from .asynciotest import *




import logging
_log = logging.getLogger(__name__)

import asyncio

from functools import partial, wraps

from . import raw
from .raw import Disconnected, RemoteError, Cancelled, Finished
from ..wrapper import Value, Type
from .._p4p import (logLevelAll, logLevelTrace, logLevelDebug,
                    logLevelInfo, logLevelWarn, logLevelError,
                    logLevelFatal, logLevelOff)

__all__ = [
    'Context',
    'Value',
    'Type',
    'RemoteError',
    'timesout',
]

if hasattr(asyncio, 'get_running_loop'): # py >=3.7
    from asyncio import get_running_loop, create_task, all_tasks
else:
    from asyncio import _get_running_loop
    from asyncio.tasks import Task

    def get_running_loop():
        ret = _get_running_loop()
        if ret is None:
            raise RuntimeError('Thread has no running event loop')
        return ret

    def create_task(coro, *, name=None):
        return get_running_loop().create_task(coro)

    def all_tasks():
        return Task.all_tasks(loop=get_running_loop())

def timesout(deftimeout=5.0):
    """Decorate a coroutine to implement an overall timeout.

    The decorated coroutine will have an additional keyword
    argument 'timeout=' which gives a timeout in seconds,
    or None to disable timeout.

    :param float deftimeout: The default timeout= for the decorated coroutine.

    It is suggested to perform one overall timeout at a high level
    rather than multiple timeouts on low-level operations. ::

        @timesout()
        async def dostuff(ctxt):
            await ctxt.put('msg', 'Working')
            A, B = await ctxt.get(['foo', 'bar'])
            await ctxt.put('bar', A+B, wait=True)
            await ctxt.put('msg', 'Done')

        async def exec():
            with Context('pva') as ctxt:
                await dostuff(ctxt, timeout=5)
    """
    def decorate(fn):
        assert asyncio.iscoroutinefunction(fn), "Place @timesout before @coroutine"

        @wraps(fn)
        async def wrapper(*args, timeout=deftimeout, **kws):
            fut = fn(*args, **kws)
            if timeout is None:
                await fut
            else:
                await asyncio.wait_for(fut, timeout=timeout)
        return wrapper
    return decorate


class Context(raw.Context):

    """
    :param str provider: A Provider name.  Try "pva" or run :py:meth:`Context.providers` for a complete list.
    :param conf dict: Configuration to pass to provider.  Depends on provider selected.
    :param bool useenv: Allow the provider to use configuration from the process environment.
    :param dict nt: Controls :ref:`unwrap`.  None uses defaults.  Set False to disable
    :param dict unwrap: Legacy :ref:`unwrap`.

    The methods of this Context will block the calling thread until completion or timeout

    The meaning, and allowed keys, of the configuration dictionary depend on the provider.

    The "pva" provider understands the following keys:

    * EPICS_PVA_ADDR_LIST
    * EPICS_PVA_AUTO_ADDR_LIST
    * EPICS_PVA_SERVER_PORT
    * EPICS_PVA_BROADCAST_PORT

    Timeout and Cancellation
    ^^^^^^^^^^^^^^^^^^^^^^^^

    All coroutines/Futures returned by Context methods can be cancelled.
    The methods of Context do not directly implement a timeout.
    Instead :py:meth:`asyncio.wait_for` should be used.
    It is suggested perform one overall timeout at a high level
    rather than multiple timeouts on low-level operations. ::

        @timesout()
        async def dostuff(ctxt):
            await ctxt.put('msg', 'Working')
            A, B = await ctxt.get(['foo', 'bar'])
            await ctxt.put('bar', A+B, wait=True)
            await ctxt.put('msg', 'Done')

        async def exec():
            with Context('pva') as ctxt:
                await dostuff(ctxt, timeout=5)
    """

    def __init__(self, provider='pva', conf=None, useenv=True, nt=None, unwrap=None):
        super(Context, self).__init__(provider, conf=conf, useenv=useenv, nt=nt, unwrap=unwrap)

    async def get(self, name, request=None):
        """Fetch current value of some number of PVs.

        :param name: A single name string or list of name strings
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.

        :returns: A p4p.Value, or list of same.  Subject to :py:ref:`unwrap`.

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values. ::

            with Context('pva') as ctxt:
                V    = await ctxt.get('pv:name')
                A, B = await ctxt.get(['pv:1', 'pv:2'])
        """
        singlepv = isinstance(name, (bytes, str))
        if singlepv:
            return (await self._get_one(name, request=request))

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)

        futs = [self._get_one(N, request=R) for N, R in zip(name, request)]

        ret = await asyncio.gather(*futs)

        return ret

    async def _get_one(self, name, request=None):
        F = asyncio.Future()

        def cb(value):
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, Exception):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(get_running_loop().call_soon_threadsafe, cb)

        op = super(Context, self).get(name, cb, request=request)

        _log.debug('get %s request=%s', name, request)
        try:
            return (await F)
        finally:
            op.close()

    async def put(self, name, values, request=None, process=None, wait=None, get=True):
        """Write a new value of some number of PVs.

        :param name: A single name string or list of name strings
        :param values: A single value, a list of values, a dict, a `Value`.  May be modified by the constructor nt= argument.
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param str process: Control remote processing.  May be 'true', 'false', 'passive', or None.
        :param bool wait: Wait for all server processing to complete.
        :param bool get: Whether to do a Get before the Put.  If True then the value passed to the builder callable
                         will be initialized with recent PV values.  eg. use this with NTEnum to find the enumeration list.

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values

        If 'wait' or 'process' is specified, then 'request' must be omitted or None. ::

            with Context('pva') as ctxt:
                await ctxt.put('pv:name', 5.0)
                await ctxt.put(['pv:1', 'pv:2'], [1.0, 2.0])
                await ctxt.put('pv:name', {'value':5})

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        if request and (process or wait is not None):
            raise ValueError("request= is mutually exclusive to process= or wait=")
        elif process or wait is not None:
            request = 'field()record[block=%s,process=%s]' % ('true' if wait else 'false', process or 'passive')

        singlepv = isinstance(name, (bytes, str))
        if singlepv:
            return (await self._put_one(name, values, request=request, get=get))

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)
        assert len(name) == len(values), (name, values)

        futs = [self._put_one(N, V, request=R, get=get) for N, V, R in zip(name, values, request)]

        await asyncio.gather(*futs)

    async def _put_one(self, name, value, request=None, get=True):
        F = asyncio.Future()

        def cb(value):
            _log.debug("put done %s %r", name, value)
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, Exception):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(get_running_loop().call_soon_threadsafe, cb)

        op = super(Context, self).put(name, cb, builder=value, request=request, get=get)

        _log.debug('put %s <- %r request=%s', name, value, request)
        try:
            value = await F
        finally:
            op.close()

    async def rpc(self, name, value, request=None):
        """Perform a Remote Procedure Call (RPC) operation

        :param str name: PV name string
        :param Value value: Arguments.  Must be Value instance
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.

        :returns: A Value.  Subject to :py:ref:`unwrap`.

        For example: ::

            uri = NTURI(['A','B'])
            with Context('pva') as ctxt:
                result = await ctxt.rpc('pv:name:add', uri.wrap('pv:name:add', 5, B=6))

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict or Value, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        F = asyncio.Future()

        def cb(value):
            if F.cancelled() or F.done():
                return  # ignore
            elif isinstance(value, Exception):
                F.set_exception(value)
            else:
                F.set_result(value)
        cb = partial(get_running_loop().call_soon_threadsafe, cb)

        op = super(Context, self).rpc(name, cb, value, request=request)

        try:
            return (await F)
        finally:
            op.close()

    def monitor(self, name, cb, request=None, notify_disconnect=False) -> "Subscription":
        """Create a callback subscription.

        :param str name: PV name string
        :param callable cb: Processing callback
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param bool notify_disconnect: In additional to Values, the callback may also be called with instances of Exception.
                                       Specifically: Disconnected , RemoteError, or Cancelled
        :returns: a :py:class:`Subscription` instance

        The callable will be invoked with one argument which is either.

        * A p4p.Value (Subject to :py:ref:`unwrap`)
        * A sub-class of Exception (Disconnected , RemoteError, or Cancelled)
        """
        assert asyncio.iscoroutinefunction(cb), "monitor callback must be coroutine"
        R = Subscription(name, cb, notify_disconnect=notify_disconnect)
        cb = partial(get_running_loop().call_soon_threadsafe, R._E.set)

        R._S = super(Context, self).monitor(name, cb, request)
        return R


class Subscription(object):

    """An active subscription.
    """

    def __init__(self, name, cb, notify_disconnect=False):
        self.name, self._S, self._cb = name, None, cb
        self._notify_disconnect = notify_disconnect

        self._run = True
        self._E = asyncio.Event()

        self._T = create_task(self._handle())

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.close()

    def close(self):
        """Begin closing subscription.
        """
        if self._S is not None:
            # after .close() self._event should never be called
            self._S.close()
            self._S = None
            self._run = False
            self._E.set()

    @property
    def done(self):
        'Has all data for this subscription been received?'
        return self._S is None or self._S.done()

    @property
    def empty(self):
        'Is data pending in event queue?'
        return self._S is None or self._S.empty()

    async def wait_closed(self):
        """Wait until subscription is closed.
        """
        assert self._S is None, "Not close()'d"
        await self._T

    async def _handle(self):
        if self._notify_disconnect:
            await self._cb(Disconnected())  # all subscriptions are inittially disconnected

        E = None
        try:
            while self._run:
                await self._E.wait()
                self._E.clear()
                _log.debug('Subscription %s wakeup', self.name)

                i = 0
                while self._run:
                    S = self._S
                    E = S.pop()
                    if E is None:
                        break

                    elif isinstance(E, Disconnected):
                        _log.debug('Subscription notify for %s with %s', self.name, E)
                        if self._notify_disconnect:
                            await self._cb(E)
                        else:
                            _log.info("Subscription disconnect %s", self.name)
                        continue

                    elif isinstance(E, RemoteError):
                        _log.debug('Subscription notify for %s with %s', self.name, E)
                        if self._notify_disconnect:
                            await self._cb(E)
                        elif isinstance(E, RemoteError):
                            _log.error("Subscription Error %s", E)
                        return

                    else:
                        await self._cb(E)

                    i = (i + 1) % 4
                    if i == 0:
                        await asyncio.sleep(0)  # Not sure how necessary.  Ensure we go to the scheduler

                    if S.done:
                        _log.debug('Subscription complete %s', self.name)
                        S.close()
                        self._S = None
                        if self._notify_disconnect:
                            E = Finished()
                            await self._cb(E)


        except asyncio.CancelledError:
            _log.debug("Cancelled Subscription: %r", self)
        except:
            _log.exception("Error processing Subscription event: %r", E)
        finally:
            if self._S is not None:
                self._S.close()
            self._S = None



"""Helper for handling NTNDArray a la. areaDetector.

Known attributes

"ColorMode"   (inner-most left, as given in NDArray.cpp, numpy.ndarray.shape is reversed from this)
 0 - Mono   [Nx, Ny]
 1 - Bayer  [Nx, Ny]
 2 - RGB1   [3, Nx, Ny]
 3 - RGB2   [Nx, 3, Ny]
 4 - RGB3   [Nx, Ny, 3]
 5 - YUV444 ?
 6 - YUV422 ??
 7 - YUV411 ???
"""

import logging
_log = logging.getLogger(__name__)

import time
import numpy

from ..wrapper import Type, Value
from .common import alarm, timeStamp, NTBase

from .scalar import ntwrappercommon


class ntndarray(ntwrappercommon, numpy.ndarray):
    """
    Augmented numpy.ndarray with additional attributes

    * .attrib    - dictionary
    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.

    Keys in the ``attrib`` dictionary may be any python which may be stored in a PVA field,
    including an arbitrary ``Value``.
    However, special handling is attempted if the provided ``Value`` appears to be an NTScalar
    or similar, in which case the .value, .alarm and .timeStamp are unpacked to the NTAttribute
    and other fields are discarded.
    """

    def __init__(self, *args, **kws):
        super(ntndarray, self).__init__(*args, **kws)
        self.attrib = {}

    def _store(self, value):
        ntwrappercommon._store(self, value)
        self.attrib = {}
        for elem in value.get('attribute', []):
            self.attrib[elem.name] = elem.value

        # we will fail if dimension[] contains None s, or if
        # the advertised shape isn't consistent with the pixel array length.
        shape = [D.size for D in value.dimension]
        shape.reverse()

        # in-place reshape!  Isn't numpy fun
        self.shape = shape or [0] # can't reshape if 0-d, so treat as 1-d if no dimensions provided

        return self


class NTNDArray(NTBase):
    """Representation of an N-dimensional array with meta-data

    Translates into `ntndarray`
    """
    Value = Value
    ntndarray = ntndarray

    # map numpy.dtype.char to .value union member name
    _code2u = {
        '?':'booleanValue',
        'b':'byteValue',
        'h':'shortValue',
        'i':'intValue',
        'l':'longValue',
        'B':'ubyteValue',
        'H':'ushortValue',
        'I':'uintValue',
        'L':'ulongValue',
        'f':'floatValue',
        'd':'doubleValue',
    }

    @classmethod
    def buildType(klass, extra=[]):
        """Build type
        """
        ret = klass._default_type
        if extra:
            L = ret.aspy()
            L.extend(extra)
            ret = Type(L, ret.getID())
        return ret

    _default_type = Type([
            ('value', ('U', None, [
                ('booleanValue', 'a?'),
                ('byteValue', 'ab'),
                ('shortValue', 'ah'),
                ('intValue', 'ai'),
                ('longValue', 'al'),
                ('ubyteValue', 'aB'),
                ('ushortValue', 'aH'),
                ('uintValue', 'aI'),
                ('ulongValue', 'aL'),
                ('floatValue', 'af'),
                ('doubleValue', 'ad'),
            ])),
            ('codec', ('S', 'codec_t', [
                ('name', 's'),
                ('parameters', 'v'),
            ])),
            ('compressedSize', 'l'),
            ('uncompressedSize', 'l'),
            ('uniqueId', 'i'),
            ('dataTimeStamp', timeStamp),
            ('alarm', alarm),
            ('timeStamp', timeStamp),
            ('dimension', ('aS', 'dimension_t', [
                ('size', 'i'),
                ('offset', 'i'),
                ('fullSize', 'i'),
                ('binning', 'i'),
                ('reverse', '?'),
            ])),
            ('attribute', ('aS', 'epics:nt/NTAttribute:1.0', [
                ('name', 's'),
                ('value', 'v'),
                ('tags', 'as'),
                ('descriptor', 's'),
                ('alarm', alarm),
                ('timeStamp', timeStamp),
                ('sourceType', 'i'),
                ('source', 's'),
            ])),
        ], id='epics:nt/NTNDArray:1.0')

    def __init__(self, **kws):
        self.type = self.buildType(**kws)

    def wrap(self, value, **kws):
        """Wrap numpy.ndarray as Value
        """
        attrib = getattr(value, 'attrib', None) or kws.pop('attrib', None) or {}

        value = numpy.asarray(value) # loses any special/augmented attributes
        dims = value.shape

        if 'ColorMode' not in attrib:
            # attempt to infer color mode from shape
            if value.ndim==2:
                attrib['ColorMode'] = 0 # gray

            elif value.ndim==3:
                for idx,dim in enumerate(reversed(dims)): # inner-most sent as left
                    if dim==3: # assume it's a color
                        attrib['ColorMode'] = 2 + idx  # 2 - RGB1, 3 - RGB2, 4 - RGB3
                        break # assume that the first is color, and any subsequent dim=3 is a thin ROI
                else:
                    raise ValueError("Unable to deduce color dimension from shape %r"%dims)

        dataSize = value.nbytes

        return self._annotate(Value(self.type, {
            'value': (self._code2u[value.dtype.char], value.flatten()),
            'compressedSize': dataSize,
            'uncompressedSize': dataSize,
            'uniqueId': 0,
            'attribute': [translateNDAttribute(K,V) for K, V in attrib.items()],
            'dimension': [{'size': N,
                           'offset': 0,
                           'fullSize': N,
                           'binning': 1,
                           'reverse': False} for N in reversed(dims)],
        }), **kws)

    @classmethod
    def unwrap(klass, value):
        """Unwrap Value as NTNDArray
        """
        V = value.value
        if V is None:
            # Union empty.  treat as zero-length char array
            V = numpy.zeros((0,), dtype=numpy.uint8)
        return V.view(klass.ntndarray)._store(value)

    def assign(self, V, py):
        """Store python value in Value
        """
        V.value = py

def translateNDAttribute(name, value):
    if isinstance(value, Value) and 'value' in value: # assume to be NT-like
        V = {
            'name': name,
            'value': value['value'],
        }
        if 'alarm' in value:
            V['alarm'] = value['alarm']
        if 'timeStamp' in value:
            V['timeStamp'] = value['timeStamp']
        return V

    return {'name': name, 'value': value}



#cython: language_level=2

from libcpp cimport bool
from libcpp.string cimport string

cdef extern from "<pvxs/data.h>" namespace "pvxs" nogil:
    cdef enum code_t "pvxs::TypeCode::code_t":
        Null "pvxs::TypeCode::Null"

        Bool "pvxs::TypeCode::Bool"
        Int8 "pvxs::TypeCode::Int8"
        Int16 "pvxs::TypeCode::Int16"
        Int32 "pvxs::TypeCode::Int32"
        Int64 "pvxs::TypeCode::Int64"
        UInt8 "pvxs::TypeCode::UInt8"
        UInt16 "pvxs::TypeCode::UInt16"
        UInt32 "pvxs::TypeCode::UInt32"
        UInt64 "pvxs::TypeCode::UInt64"
        Float32 "pvxs::TypeCode::Float32"
        Float64 "pvxs::TypeCode::Float64"
        String "pvxs::TypeCode::String"
        Struct "pvxs::TypeCode::Struct"
        Union "pvxs::TypeCode::Union"
        Any "pvxs::TypeCode::Any"

        BoolA "pvxs::TypeCode::BoolA"
        Int8A "pvxs::TypeCode::Int8A"
        Int16A "pvxs::TypeCode::Int16A"
        Int32A "pvxs::TypeCode::Int32A"
        Int64A "pvxs::TypeCode::Int64A"
        UInt8A "pvxs::TypeCode::UInt8A"
        UInt16A "pvxs::TypeCode::UInt16A"
        UInt32A "pvxs::TypeCode::UInt32A"
        UInt64A "pvxs::TypeCode::UInt64A"
        Float32A "pvxs::TypeCode::Float32A"
        Float64A "pvxs::TypeCode::Float64A"
        StringA "pvxs::TypeCode::StringA"
        StructA "pvxs::TypeCode::StructA"
        UnionA "pvxs::TypeCode::UnionA"
        AnyA "pvxs::TypeCode::AnyA"

    cdef enum StoreType:
        SNull "pvxs::StoreType::Null"
        SBool "pvxs::StoreType::Bool"
        SUInteger "pvxs::StoreType::UInteger"
        SInteger "pvxs::StoreType::Integer"
        SReal "pvxs::StoreType::Real"
        SString "pvxs::StoreType::String"
        SCompound "pvxs::StoreType::Compound"
        SArray "pvxs::StoreType::Array"

    cdef enum Kind:
        KBool     "Kind::Bool"
        KInteger  "Kind::Integer"
        Real      "Kind::Real"
        KString   "Kind::String"
        KCompound "Kind::Compound"
        KNull     "Kind::Null"

    cdef cppclass TypeCode:
        code_t code

        TypeCode()
        TypeCode(unsigned)
        TypeCode(code_t)

        bool valid() const
        #Kind kind() const
        unsigned order() const
        unsigned size() const
        bool isunsigned() const
        bool isarray() const

        StoreType storeAs() const

        TypeCode arrayOf() const
        TypeCode scalarOf() const

        const char* name() const

    cdef cppclass Member:
        Member()
        Member(TypeCode, const string&) except+
        #Member[IT](TypeCode, const string&, const IT&) except+
        #Member[IT](TypeCode, const string&, const string&, const IT&) except+

        void addChild(const Member& mem) except+


    cdef cppclass TypeDef:
        TypeDef()
        TypeDef(const TypeDef&)
        TypeDef(TypeDef&&)

        TypeDef(const Value&) except+

        TypeDef(TypeCode)
        # Cython has a problem with parameterized constructors
        #TypeDef[IT](TypeCode, const string&, const IT&) except+

        Member as(const string&) except+

        Value create() except+

    cdef cppclass Value:
        Value()
        Value(const Value&)
        Value(Value&&)

        Value cloneEmpty() except+
        Value clone() except+

        Value& assign(const Value&) except+

        Value allocMember() except+

        bool valid() const
        bool operator bool() const

        bool isMarked(bool, bool) const
        Value ifMarked(bool, bool) const

        void mark(bool) except+
        void unmark(bool, bool) except+

        TypeCode type() const
        StoreType storageType() const
        const string& id() except+
        bool idStartsWith(const string&) except+

        bool equalInst(const Value&)

        const string& nameOf(const Value&) except+

        void copyOut(void *ptr, StoreType type) except+
        bool tryCopyOut(void *ptr, StoreType type) except+
        void copyIn(const void *ptr, StoreType type) except+
        bool tryCopyIn(const void *ptr, StoreType type) except+

        V as[V]() except+
        bool as[V](V&) except+

        bool tryFrom[V](const V&) except+
        void _from "from" [V](const V&) except+

        Value& update[K,V](K, const V&) except+

        Value operator[](const string&) except+
        Value lookup(const string&) except+ KeyError

        size_t nmembers() const

        cppclass IAll:
            cppclass iterator:
                Value& operator*()
                iterator operator++()
                iterator operator--()
                iterator operator+(size_type)
                iterator operator-(size_type)
                #difference_type operator-(iterator)
                bint operator==(iterator)
                bint operator!=(iterator)
                bint operator<(iterator)
                bint operator>(iterator)
                bint operator<=(iterator)
                bint operator>=(iterator)
            iterator begin()
            iterator end()

        cppclass IChildren:
            cppclass iterator:
                Value& operator*()
                iterator operator++()
                iterator operator--()
                iterator operator+(size_type)
                iterator operator-(size_type)
                #difference_type operator-(iterator)
                bint operator==(iterator)
                bint operator!=(iterator)
                bint operator<(iterator)
                bint operator>(iterator)
                bint operator<=(iterator)
                bint operator>=(iterator)
            iterator begin()
            iterator end()

        cppclass IMarked:
            cppclass iterator:
                Value& operator*()
                iterator operator++()
                iterator operator--()
                iterator operator+(size_type)
                iterator operator-(size_type)
                #difference_type operator-(iterator)
                bint operator==(iterator)
                bint operator!=(iterator)
                bint operator<(iterator)
                bint operator>(iterator)
                bint operator<=(iterator)
                bint operator>=(iterator)
            iterator begin()
            iterator end()

        IAll iall() except+
        IChildren ichildren() except+
        IMarked imarked() except+



from __future__ import print_function

import logging
import unittest
import random
import weakref
import sys
import gc
import inspect
import threading

try:
    from Queue import Queue, Full, Empty
except ImportError:
    from queue import Queue, Full, Empty

from ..wrapper import Value, Type
from ..client.thread import Context, Disconnected, TimeoutError, RemoteError
from ..server import Server, StaticProvider
from ..server.thread import SharedPV, _defaultWorkQueue
from ..util import WorkQueue
from ..nt import NTScalar, NTURI
from .utils import RefTestCase

_log = logging.getLogger(__name__)

class TestGPM(RefTestCase):
    maxDiff = 1000
    timeout = 1.0

    class Times2Handler(object):

        def put(self, pv, op):
            V = op.value()
            if V.raw.changed('value'):
                if V < 0:
                    op.done(error="Must be non-negative")
                V = V * 2
                pv.post(V)
            op.done()

    def setUp(self):
        # gc.set_debug(gc.DEBUG_LEAK)
        super(TestGPM, self).setUp()

        self.pv = SharedPV(handler=self.Times2Handler(), nt=NTScalar('d'))
        self.pv2 = SharedPV(handler=self.Times2Handler(), nt=NTScalar('d'), initial=42.0)
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)
        self.sprov.add('bar', self.pv2)

        self.server = Server(providers=[self.sprov], isolate=True)
        _log.debug('Server Conf: %s', self.server.conf())

    def tearDown(self):
        self.server.stop()
        _defaultWorkQueue.sync()
        #self.pv._handler._pv = None
        R = [weakref.ref(r) for r in (self.server, self.sprov, self.pv, self.pv._whandler, self.pv._handler)]
        r = None
        del self.server
        del self.sprov
        del self.pv
        del self.pv2
        gc.collect()
        R = [r() for r in R]
        self.assertListEqual(R, [None] * len(R))
        super(TestGPM, self).tearDown()

    def testGet(self):
        with Context('pva', conf=self.server.conf(), useenv=False) as ctxt:
            _log.debug('Client conf: %s', ctxt.conf())
            # PV not yet opened
            self.assertRaises(TimeoutError, ctxt.get, 'foo', timeout=0.1)

            self.pv.open(1.0)

            V = ctxt.get('foo')
            self.assertEqual(V, 1.0)
            self.assertTrue(V.raw.changed('value'))

            self.assertEqual(ctxt.get(['foo', 'bar']), [1.0, 42.0])

        C = weakref.ref(ctxt)
        del ctxt
        gc.collect()
        self.assertIsNone(C())

    def testPutGet(self):
        with Context('pva', conf=self.server.conf(), useenv=False) as ctxt:

            self.pv.open(1.0)

            V = ctxt.get('foo')
            self.assertEqual(V, 1.0)

            ctxt.put('foo', 5)

            V = ctxt.get('foo')
            self.assertEqual(V, 10.0)

            ctxt.put(['foo', 'bar'], [5, 6])

            self.assertEqual(ctxt.get(['foo', 'bar']), [5 * 2, 6 * 2])

        C = weakref.ref(ctxt)
        del ctxt
        gc.collect()
        self.assertIsNone(C())

    def testMonitor(self):
        with Context('pva', conf=self.server.conf(), useenv=False) as ctxt:

            self.pv.open(1.0)

            Q = Queue(maxsize=4)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)

            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 1.0)

            ctxt.put('foo', 4)

            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 8.0)

            self.pv.close()

            V = Q.get(timeout=self.timeout)
            self.assertIsInstance(V, Disconnected)

            self.pv.open(3.0)
            ctxt.hurryUp()

            V = Q.get(timeout=self.timeout)
            self.assertEqual(V, 3.0)

        C = weakref.ref(ctxt)
        del ctxt
        del sub
        del Q
        gc.collect()
        self.assertIsNone(C())

class TestRPC(RefTestCase):
    maxDiff = 1000
    timeout = 1.0
    openclose = False

    class Handler:
        def __init__(self, openclose):
            self.openclose = openclose
        def onFirstConnect(self, pv):
            _log.debug("onFirstConnect")
            if self.openclose:
                pv.open(0)

        def onLastDisconnect(self, pv):
            _log.debug("onLastDisconnect")
            if self.openclose:
                pv.close()

        def rpc(self, pv, op):
            V = op.value()
            if V.get('query.oops'):
                op.done(error='oops')
            elif V.get('query.null'):
                op.done()
            else:
                op.done(NTScalar('i').wrap(42))

    def setUp(self):
        super(TestRPC, self).setUp()

        self.pv = SharedPV(nt=NTScalar('i'), handler=self.Handler(self.openclose))
        self.provider = StaticProvider("serverend")
        self.provider.add('foo', self.pv)

    def tearDown(self):
        self.pv.close(sync=True, timeout=self.timeout)
        self.traceme(self.pv)
        self.traceme(self.provider)
        del self.pv
        del self.provider
        _defaultWorkQueue.sync()
            
        super(TestRPC, self).tearDown()

    def test_rpc(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                args = NTURI([
                    ('lhs', 'd'),
                    ('rhs', 'd'),
                ])

                # self.pv not open()'d
                ret = C.rpc('foo', args.wrap('foo', kws={'lhs':1, 'rhs':2}))
                _log.debug("RET %s", ret)
                self.assertEqual(ret, 42)

                ret = C.rpc('foo', None)
                _log.debug("RET %s", ret)
                self.assertEqual(ret, 42)

    def test_rpc_null(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                args = NTURI([
                    ('null', '?'),
                ])

                # self.pv not open()'d
                ret = C.rpc('foo', args.wrap('foo', kws={'null':True}))
                _log.debug("RET %s", ret)
                self.assertIsNone(ret)

    def test_rpc_error(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False) as C:

                args = NTURI([
                    ('oops', '?'),
                ])

                with self.assertRaisesRegexp(RemoteError, 'oops'):
                    ret = C.rpc('foo', args.wrap('foo', kws={'oops':True}))

class TestRPC2(TestRPC):
    openclose = True

class TestPVRequestMask(RefTestCase):
    maxDiff = 1000
    timeout = 1.0
    mode = 'Mask'

    class Handler(object):
        def put(self, pv, op):
            val = op.value()
            _log.debug("Putting %s %s", val.raw.changedSet(), str(val.raw))
            pv.post(op.value())
            op.done()

    def setUp(self):
        # gc.set_debug(gc.DEBUG_LEAK)
        super(TestPVRequestMask, self).setUp()

        self.pv = SharedPV(handler=self.Handler(),
                           nt=NTScalar('d'),
                           initial=1.0,
                           options={'mapperMode':self.mode})
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)

        self.server = Server(providers=[self.sprov], isolate=True)

    def tearDown(self):
        self.server.stop()
        _defaultWorkQueue.sync()
        self.pv._handler._pv = None
        R = [weakref.ref(r) for r in (self.server, self.sprov, self.pv, self.pv._whandler, self.pv._handler)]
        r = None
        del self.server
        del self.sprov
        del self.pv
        gc.collect()
        R = [r() for r in R]
        self.assertListEqual(R, [None] * len(R))
        super(TestPVRequestMask, self).tearDown()

    def testGetPut(self):
        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            V = ctxt.get('foo', request='value')

            self.assertEqual(V.value, 1.0)

            self.assertTrue(V.changed('value'))

            if self.mode=='Mask':
                self.assertSetEqual(V.changedSet(), {'value'})
            else:
                self.assertListEqual(V.keys(), ['value'])

            ctxt.put('foo', {'value':2.0}, request='value')

            V = ctxt.get('foo', request='value')

            self.assertEqual(V.value, 2.0)

    def testMonitor(self):
        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:

            Q = Queue(maxsize=4)
            sub = ctxt.monitor('foo', Q.put, request='value')

            V = Q.get(timeout=self.timeout)
            self.assertEqual(V.value, 1.0)

            if self.mode=='Mask':
                self.assertSetEqual(V.changedSet(), {'value'})
            else:
                self.assertListEqual(V.keys(), ['value'])

            ctxt.put('foo', {'alarm.severity':1}) # should be dropped

            ctxt.put('foo', {'value':3.0}, request='value')

            V = Q.get(timeout=self.timeout)
            self.assertEqual(V.value, 3.0)

            if self.mode=='Mask':
                self.assertSetEqual(V.changedSet(), {'value'})
            else:
                self.assertListEqual(V.keys(), ['value'])


#class TestPVRequestSlice(TestPVRequestMask):
#    mode = 'Slice'

class TestFirstLast(RefTestCase):
    maxDiff = 1000
    timeout = 1.0
    mode = 'Mask'

    class Handler:
        def __init__(self):
            self.evtC = threading.Event()
            self.evtD = threading.Event()
            self.conn = None
        def onFirstConnect(self, pv):
            _log.debug("onFirstConnect")
            self.conn = True
            self.evtC.set()
        def onLastDisconnect(self, pv):
            _log.debug("onLastDisconnect")
            self.conn = False
            self.evtD.set()

    def setUp(self):
        super(TestFirstLast, self).setUp()

        self.H = self.Handler()
        self.pv = SharedPV(handler=self.H,
                           nt=NTScalar('d'),
                           options={'mapperMode':self.mode})
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)

        self.server = Server(providers=[self.sprov], isolate=True)

    def tearDown(self):
        self.server.stop()
        _defaultWorkQueue.sync()
        del self.server
        del self.sprov
        del self.pv
        del self.H
        super(TestFirstLast, self).tearDown()

    def testClientDisconn(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = Queue(maxsize=4)
            with ctxt.monitor('foo', Q.put, notify_disconnect=True):

                self.assertIsInstance(Q.get(timeout=self.timeout), Disconnected)
                self.assertIsInstance(Q.get(timeout=self.timeout), Value) # initial update

                _log.debug('TEST')
                self.H.evtC.wait(self.timeout) # onFirstConnect()
                self.assertTrue(self.H.conn)

        self.H.evtD.wait(self.timeout) # onLastDisconnect()
        _log.debug('SHUTDOWN')
        self.assertFalse(self.H.conn)

    def testServerShutdown(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = Queue(maxsize=4)
            with ctxt.monitor('foo', Q.put, notify_disconnect=True):

                Q.get(timeout=self.timeout) # initial update

                _log.debug('TEST')
                self.H.evtC.wait(self.timeout) # onFirstConnect()
                self.assertIs(self.H.conn, True)

                self.server.stop()

                self.H.evtD.wait(self.timeout) # onLastDisconnect()
                _log.debug('SHUTDOWN')
                self.assertIs(self.H.conn, False)

    def testPVClose(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
            Q = Queue(maxsize=4)
            with ctxt.monitor('foo', Q.put, notify_disconnect=True):

                Q.get(timeout=self.timeout) # initial update

                _log.debug('TEST')
                self.H.evtC.wait(self.timeout) # onFirstConnect()
                self.assertTrue(self.H.conn)

                self.pv.close(destroy=True, sync=True, timeout=self.timeout) # onLastDisconnect()

                _log.debug('CLOSE')
                self.assertFalse(self.H.conn)



#cython: language_level=2

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.map cimport map

from .data cimport Value
from .source cimport Source

cdef extern from "<pvxs/sharedpv.h>" namespace "pvxs::server" nogil:
    cdef cppclass SharedPV:
        @staticmethod
        SharedPV buildMailbox() except+
        @staticmethod
        SharedPV buildReadonly() except+

        SharedPV()

        #void attach(std::unique_ptr<ChannelControl>&& op);
        #void onFirstConnect(std::function<void()>&& fn);
        #void onLastDisconnect(std::function<void()>&& fn);
        #void onPut(std::function<void(SharedPV&, std::unique_ptr<ExecOp>&&, Value&&)>&& fn);
        #void onRPC(std::function<void(SharedPV&, std::unique_ptr<ExecOp>&&, Value&&)>&& fn);

        void open(const Value& initial) except+
        bool isOpen() except+
        void close() except+

        void post(const Value& val) except+
        void fetch(Value& val) except+
        Value fetch() except+

    cdef cppclass StaticSource:
        @staticmethod
        StaticSource build() except+;

        StaticSource()

        shared_ptr[Source] source() except+

        void close() except+
        StaticSource& add(const string& name, const SharedPV& pv) except+
        StaticSource& remove(const string& name) except+

        map[string, SharedPV] list() except+




import logging
_log = logging.getLogger(__name__)

from functools import partial
from threading import Event

from ..util import _defaultWorkQueue
from .raw import SharedPV as _SharedPV, Handler
from ..client.raw import RemoteError

__all__ = (
    'SharedPV',
    'Handler',
)

def _on_queue(op, M, *args):
    try:
        M(*args)
        return
    except RemoteError as e:
        err = e
    except Exception as e:
        _log.exception("Unexpected")
        err = e
    if op is not None:
        op.done(error=str(err))


class SharedPV(_SharedPV):

    """Shared state Process Variable.  Callback based implementation.

    .. note:: if initial=None, the PV is initially **closed** and
              must be :py:meth:`open()`'d before any access is possible.

    :param handler: A object which will receive callbacks when eg. a Put operation is requested.
                    May be omitted if the decorator syntax is used.
    :param Value initial: An initial Value for this PV.  If omitted, :py:meth:`open` s must be called before client access is possible.
    :param nt: An object with methods wrap() and unwrap().  eg :py:class:`p4p.nt.NTScalar`.
    :param callable wrap: As an alternative to providing 'nt=', A callable to transform Values passed to open() and post().
    :param callable unwrap: As an alternative to providing 'nt=', A callable to transform Values returned Operations in Put/RPC handlers.
    :param WorkQueue queue: The threaded :py:class:`WorkQueue` on which handlers will be run.
    :param dict options: A dictionary of configuration options.

    Creating a PV in the open state, with no handler for Put or RPC (attempts will error). ::

        from p4p.nt import NTScalar
        pv = SharedPV(nt=NTScalar('d'), value=0.0)
        # ... later
        pv.post(1.0)

    The full form of a handler object is: ::

        class MyHandler:
            def put(self, op):
                pass
            def rpc(self, op):
                pass
            def onFirstConnect(self): # may be omitted
                pass
            def onLastDisconnect(self): # may be omitted
                pass
        pv = SharedPV(MyHandler())

    Alternatively, decorators may be used. ::

        pv = SharedPV()
        @pv.put
        def onPut(pv, op):
            pass

    """

    def __init__(self, queue=None, **kws):
        _SharedPV.__init__(self, **kws)
        self._queue = queue or _defaultWorkQueue()
        self._disconnected = Event()
        self._disconnected.set()

    def _exec(self, op, M, *args):
        self._queue.push(partial(_on_queue, op, M, *args))

    def _onFirstConnect(self, _junk):
        self._disconnected.clear()

    def _onLastDisconnect(self, _junk):
        self._disconnected.set()

    def close(self, destroy=False, sync=False, timeout=None):
        """Close PV, disconnecting any clients.

        :param bool destroy: Indicate "permanent" closure.  Current clients will not see subsequent open().
        :param bool sync: When block until any pending onLastDisconnect() is delivered (timeout applies).
        :param float timeout: Applies only when sync=True.  None for no timeout, otherwise a non-negative floating point value.

        close() with destory=True or sync=True will not prevent clients from re-connecting.
        New clients may prevent sync=True from succeeding.
        Prevent reconnection by __first__ stopping the Server, removing with :py:meth:`StaticProvider.remove()`,
        or preventing a :py:class:`DynamicProvider` from making new channels to this SharedPV.
        """
        _SharedPV.close(self, destroy)
        if sync:
            # TODO: still not syncing PVA workers...
            self._queue.sync()
            self._disconnected.wait()
